
#line 1 "lib/eolian/eo_lexer.rl"
#include <stdio.h>
#include <stdlib.h>
#include <Eina.h>

#include "Eolian.h"
#
#include "eo_lexer.h"
#include "eolian_database.h"

static int _eo_tokenizer_log_dom = -1;
#ifdef CRITICAL
#undef CRITICAL
#endif
#define CRITICAL(...) EINA_LOG_DOM_CRIT(_eo_tokenizer_log_dom, __VA_ARGS__)

#ifdef ERR
#undef ERR
#endif
#define ERR(...) EINA_LOG_DOM_ERR(_eo_tokenizer_log_dom, __VA_ARGS__)

#ifdef WRN
#undef WRN
#endif
#define WRN(...) EINA_LOG_DOM_WARN(_eo_tokenizer_log_dom, __VA_ARGS__)

#define INF_ENABLED EINA_FALSE
#ifdef INF
#undef INF
#endif
#define INF(...) if (INF_ENABLED) EINA_LOG_DOM_INFO(_eo_tokenizer_log_dom, __VA_ARGS__)

#define DBG_ENABLED EINA_FALSE
#ifdef DBG
#undef DBG
#endif
#define DBG(...) if (DBG_ENABLED) EINA_LOG_DOM_DBG(_eo_tokenizer_log_dom, __VA_ARGS__)

#define FUNC_PUBLIC 0
#define FUNC_PROTECTED 1

static int _init_counter = 0;

int
eo_tokenizer_init()
{
   if (!_init_counter)
     {
        eina_init();
        eina_log_color_disable_set(EINA_FALSE);
        _eo_tokenizer_log_dom = eina_log_domain_register("eo_toknz", EINA_COLOR_CYAN);
     }
   return _init_counter++;
}

int
eo_tokenizer_shutdown()
{
   if (_init_counter <= 0) return 0;
   _init_counter--;
   if (!_init_counter)
     {
        eina_log_domain_unregister(_eo_tokenizer_log_dom);
        _eo_tokenizer_log_dom = -1;
        eina_shutdown();
     }
   return _init_counter;
}

static void
_eo_tokenizer_abort(Eo_Tokenizer *toknz,
                    const char *file, const char* fct, int line,
                    const char *fmt, ...)
{
   va_list ap;
   va_start (ap, fmt);
   eina_log_vprint(_eo_tokenizer_log_dom, EINA_LOG_LEVEL_ERR,
                   file, fct, line, fmt, ap);
   va_end(ap);
   fprintf(stderr, "File:%s\n toknz[%d] n:%d l:%d p:%d pe:%d ts:%s te:%s act:%d\n",
          toknz->source,
          toknz->cs, toknz->current_nesting, toknz->current_line,
          (int)(toknz->p - toknz->buf), (int)(toknz->pe - toknz->buf),
          toknz->ts, toknz->te, toknz->act);
   exit(EXIT_FAILURE);
}
#define ABORT(toknz, ...) \
   _eo_tokenizer_abort(toknz, __FILE__, __FUNCTION__, __LINE__, __VA_ARGS__);

static void _eo_tokenizer_normalize_buf(char *buf)
{
   int c;
   char *s, *d;
   Eina_Bool in_space = EINA_TRUE;
   Eina_Bool in_newline = EINA_FALSE;

   /* ' '+ -> ' '
    * '\n' ' '* '*' ' '* -> '\n'
    */
   for (s = buf, d = buf; *s != '\0'; s++)
     {
        c = *s;
        *d = c;

        if (!in_space || (c != ' '))
          d++;

        if (c == ' ')
          in_space = EINA_TRUE;
        else
          in_space = EINA_FALSE;

        if (c == '\n')
          {
             in_newline = EINA_TRUE;
             in_space = EINA_TRUE;
          }
        else if (in_newline && c == '*' )
          {
             in_space = EINA_TRUE;
             in_newline = EINA_FALSE;
             d--;
          }
     }
   /* ' '+$ -> $ */
   d--;
   while (*d == ' ') d--;
   d++;
   if (d < buf) return;
   *d = '\0';
}

static const char*
_eo_tokenizer_token_get(Eo_Tokenizer *toknz, char *p)
{
   if (toknz->saved.tok == NULL) ABORT(toknz, "toknz->saved.tok is NULL");
   char d[BUFSIZE];
   int l = (p - toknz->saved.tok);
   memcpy(d, toknz->saved.tok, l);
   d[l] = '\0';
   _eo_tokenizer_normalize_buf(d);
   toknz->saved.tok = NULL;
   DBG("token : >%s<", d);
   return eina_stringshare_add(d);
}

static Eo_Class_Def*
_eo_tokenizer_class_get(Eo_Tokenizer *toknz, char *p)
{
   Eo_Class_Def *kls = calloc(1, sizeof(Eo_Class_Def));
   if (kls == NULL) ABORT(toknz, "calloc Eo_Class_Def failure");

   kls->name = _eo_tokenizer_token_get(toknz, p);

   return kls;
}

static Eo_Property_Def*
_eo_tokenizer_property_get(Eo_Tokenizer *toknz, char *p)
{
   Eo_Property_Def *prop = calloc(1, sizeof(Eo_Property_Def));
   if (prop == NULL) ABORT(toknz, "calloc Eo_Property_Def failure");

   prop->name = _eo_tokenizer_token_get(toknz, p);
   prop->scope = toknz->tmp.fscope;
   toknz->tmp.fscope = FUNC_PUBLIC;

   return prop;
}

static Eo_Method_Def*
_eo_tokenizer_method_get(Eo_Tokenizer *toknz, char *p)
{
   Eo_Method_Def *meth = calloc(1, sizeof(Eo_Method_Def));
   if (meth == NULL) ABORT(toknz, "calloc Eo_Method_Def failure");

   meth->name = _eo_tokenizer_token_get(toknz, p);
   meth->scope = toknz->tmp.fscope;
   toknz->tmp.fscope = FUNC_PUBLIC;

   return meth;
}

static int
_eo_tokenizer_scope_get(Eo_Tokenizer *toknz, EINA_UNUSED char *p)
{
   if (!strncmp(toknz->saved.tok, "protected ", 10))
     return FUNC_PROTECTED;

   return FUNC_PUBLIC;
}

static Eo_Param_Def*
_eo_tokenizer_param_get(Eo_Tokenizer *toknz, char *p)
{
   char *s;

   Eo_Param_Def *param = calloc(1, sizeof(Eo_Param_Def));
   if (param == NULL) ABORT(toknz, "calloc Eo_Param_Def failure");

   /* The next code part tries to identify the different tags of the
      parameter.
      First, we set the ';' to '\0', to search only inside this section.
      We then strstr the different tags and if found, we update the internal
      flag and clear the zone of the text. In this way, during the
      determination of the type/variable, we will not be disturbed by the
      flags.
      We have to put back the ';' at the end.
    */
   *p = '\0';
   s = strstr(toknz->saved.tok, "@nonull");
   if (s)
     {
        param->nonull = EINA_TRUE;
        memset(s, ' ', 7);
     }
   *p = ';';
   s = p - 1; /* Don't look at the character ';' */
   /* Remove any space between the param name and ';'/@nonull
    * This loop fixes the case where "char *name ;" becomes the type of the param.
    */
   while (*s == ' ') s--;
   for (; s >= toknz->saved.tok; s--)
     {
        if ((*s == ' ') || (*s == '*'))
          break;
     }

   if (s == toknz->saved.tok)
     ABORT(toknz, "wrong parameter: %s", _eo_tokenizer_token_get(toknz, p));
   s++;

   param->way = PARAM_IN;
   if (strncmp(toknz->saved.tok, "@in ", 4) == 0)
     {
        toknz->saved.tok += 3;
        param->way = PARAM_IN;
     }
   else if (strncmp(toknz->saved.tok, "@out ", 5) == 0)
     {
        toknz->saved.tok += 4;
        param->way = PARAM_OUT;
     }
   else if (strncmp(toknz->saved.tok, "@inout ", 7) == 0)
     {
        toknz->saved.tok += 6;
        param->way = PARAM_INOUT;
     }

   param->type = _eo_tokenizer_token_get(toknz, s);

   toknz->saved.tok = s;
   param->name = _eo_tokenizer_token_get(toknz, p);

   return param;
}

static Eo_Ret_Def*
_eo_tokenizer_return_get(Eo_Tokenizer *toknz, char *p)
{
   char *s;

   Eo_Ret_Def *ret = calloc(1, sizeof(Eo_Ret_Def));
   if (ret == NULL) ABORT(toknz, "calloc Eo_Ret_Def failure");

   *p = '\0';
   s = strstr(toknz->saved.tok, "@warn_unused");
   if (s)
     {
        ret->warn_unused = EINA_TRUE;
        memset(s, ' ', 12);
     }
   s = strchr(toknz->saved.tok, '(');
   if (s)
     {
        char *ret_val;
        char *end = strchr(s, ')');
        if (!end)
           ABORT(toknz, "wrong syntax (missing ')'): %s",
                 _eo_tokenizer_token_get(toknz, p));
        /* Current values in s and end have to be changed to ' ' to not disturb the next steps (type extraction) */
        *s++ = ' ';
        while (*s == ' ') s++;
        *end-- = ' ';
        while (end > s && *end == ' ') end--;
        if (end < s)
           ABORT(toknz, "empty default return value: %s",
                 _eo_tokenizer_token_get(toknz, p));
        ret_val = malloc(end - s + 2); /* string + '\0' */
        memcpy(ret_val, s, end - s + 1);
        ret_val[end - s + 1] = '\0';
        ret->dflt_ret_val = ret_val;
        memset(s, ' ', end - s + 1);
     }
   *p = ';';
   s = p - 1; /* Don't look at the character ';' */
   /* Remove any space between the param name and ';'
    * This loop fixes the case where "char *name ;" becomes the type of the param.
    */
   while (*s == ' ') s--;

   if (s == toknz->saved.tok)
     ABORT(toknz, "wrong parameter: %s", _eo_tokenizer_token_get(toknz, p));
   s++;

   ret->type = _eo_tokenizer_token_get(toknz, s);

   return ret;
}

static Eo_Accessor_Param*
_eo_tokenizer_accessor_param_get(Eo_Tokenizer *toknz, char *p)
{
   Eo_Accessor_Param *param = calloc(1, sizeof(Eo_Accessor_Param));
   if (param == NULL) ABORT(toknz, "calloc Eo_Accessor_Param failure");

   /* Remove the colon and spaces - we just need the param name */
   while (*p != ':') p--;
   p--;
   while (*p == ' ') p--;
   param->name = _eo_tokenizer_token_get(toknz, p + 1);

   return param;
}

static Eo_Accessor_Def *
_eo_tokenizer_accessor_get(Eo_Tokenizer *toknz, Eo_Accessor_Type type)
{
   Eo_Accessor_Def *accessor = calloc(1, sizeof(Eo_Accessor_Def));
   if (accessor == NULL) ABORT(toknz, "calloc Eo_Accessor_Def failure");

   accessor->type = type;

   return accessor;
}

static Eo_Event_Def*
_eo_tokenizer_event_get(Eo_Tokenizer *toknz, char *p)
{
   Eo_Event_Def *sgn = calloc(1, sizeof(Eo_Event_Def));
   if (sgn == NULL) ABORT(toknz, "calloc Eo_Event_Def failure");

   sgn->name = _eo_tokenizer_token_get(toknz, p);

   return sgn;
}

static Eo_Implement_Def*
_eo_tokenizer_implement_get(Eo_Tokenizer *toknz, char *p)
{
   Eo_Implement_Def *impl = calloc(1, sizeof(Eo_Implement_Def));
   if (impl == NULL) ABORT(toknz, "calloc Eo_Implement_Def failure");

   impl->meth_name = _eo_tokenizer_token_get(toknz, p);

   return impl;
}


#line 442 "lib/eolian/eo_lexer.rl"



#line 366 "lib/eolian/eo_lexer.c"
static const char _eo_tokenizer_actions[] = {
	0, 1, 0, 1, 2, 1, 7, 1, 
	11, 1, 17, 1, 19, 1, 20, 1, 
	21, 1, 22, 1, 23, 1, 24, 1, 
	25, 1, 26, 1, 27, 1, 28, 1, 
	29, 1, 30, 1, 31, 1, 32, 1, 
	33, 1, 34, 1, 35, 1, 36, 1, 
	37, 1, 40, 1, 41, 1, 42, 1, 
	43, 1, 44, 1, 45, 1, 46, 1, 
	47, 1, 48, 1, 49, 1, 50, 1, 
	51, 1, 52, 1, 53, 1, 54, 1, 
	55, 1, 56, 1, 57, 1, 58, 1, 
	59, 1, 60, 1, 61, 1, 62, 1, 
	63, 1, 64, 1, 65, 1, 66, 1, 
	67, 1, 68, 1, 69, 1, 70, 1, 
	71, 1, 72, 1, 73, 1, 74, 1, 
	75, 1, 76, 1, 77, 1, 78, 1, 
	81, 1, 82, 1, 83, 1, 84, 1, 
	85, 1, 86, 1, 87, 1, 88, 1, 
	89, 1, 90, 1, 91, 1, 92, 1, 
	93, 1, 94, 1, 95, 1, 96, 1, 
	97, 1, 98, 1, 99, 1, 100, 1, 
	101, 1, 102, 1, 103, 1, 104, 1, 
	105, 1, 106, 1, 107, 1, 108, 1, 
	109, 1, 110, 1, 111, 1, 112, 1, 
	113, 1, 114, 1, 115, 1, 116, 1, 
	117, 1, 118, 1, 119, 1, 120, 2, 
	0, 37, 2, 0, 48, 2, 0, 57, 
	2, 0, 68, 2, 0, 77, 2, 0, 
	89, 2, 0, 98, 2, 0, 115, 2, 
	5, 43, 2, 6, 38, 2, 7, 2, 
	2, 8, 39, 2, 9, 52, 2, 11, 
	0, 2, 11, 69, 2, 14, 84, 2, 
	15, 79, 2, 16, 80, 2, 17, 0, 
	2, 17, 90, 2, 19, 0, 2, 20, 
	0, 2, 20, 116, 2, 21, 0, 2, 
	22, 0, 2, 23, 0, 2, 23, 2, 
	2, 27, 0, 2, 32, 0, 2, 32, 
	116, 2, 35, 1, 2, 35, 2, 2, 
	35, 4, 2, 35, 10, 2, 35, 13, 
	3, 3, 12, 2, 3, 3, 18, 2
	
};

static const short _eo_tokenizer_key_offsets[] = {
	0, 2, 5, 6, 7, 8, 9, 10, 
	11, 12, 15, 22, 37, 41, 49, 54, 
	58, 70, 75, 82, 83, 84, 85, 88, 
	89, 90, 91, 92, 93, 94, 95, 98, 
	99, 100, 101, 104, 107, 109, 112, 123, 
	125, 128, 129, 133, 140, 147, 159, 170, 
	182, 194, 206, 218, 229, 237, 245, 257, 
	269, 281, 293, 304, 321, 338, 355, 372, 
	376, 377, 378, 388, 390, 393, 409, 411, 
	414, 415, 429, 433, 434, 435, 445, 447, 
	450, 452, 455, 456, 457, 461, 462, 463, 
	467, 468, 472, 473, 474, 475, 476, 480, 
	482, 485, 486, 490, 501, 511, 523, 535, 
	547, 559, 571, 583, 595, 605, 615, 628, 
	640, 652, 664, 676, 688, 700, 712, 719, 
	731, 743, 755, 767, 779, 791, 803, 815, 
	818, 820, 823, 834, 836, 839, 840, 841, 
	842, 843, 844, 845, 846, 847, 848, 851, 
	858, 866, 867, 868, 869, 870, 874, 875, 
	876, 877, 878, 881, 897, 913, 930, 947, 
	951, 952, 953, 963, 965, 968, 984, 986, 
	989, 990, 994, 1005, 1015, 1027, 1039, 1051, 
	1063, 1075, 1087, 1099, 1109, 1119, 1132, 1144, 
	1156, 1168, 1180, 1192, 1204, 1216, 1223, 1235, 
	1247, 1259, 1271, 1283, 1295, 1307, 1319, 1322, 
	1324, 1327, 1338, 1340, 1343, 1344, 1345, 1346, 
	1347, 1348, 1349, 1350, 1351, 1352, 1353, 1354, 
	1358, 1359, 1360, 1364, 1371, 1379, 1380, 1381, 
	1382, 1383, 1384, 1385, 1386, 1390, 1397, 1405, 
	1406, 1407, 1408, 1409, 1413, 1421, 1437, 1441, 
	1453, 1454, 1455, 1465, 1467, 1470, 1478, 1486, 
	1494, 1506, 1510, 1511, 1512, 1513, 1514, 1515, 
	1516, 1517, 1518, 1522, 1530, 1542, 1546, 1547, 
	1548, 1549, 1550, 1551, 1552, 1553, 1554, 1555, 
	1556, 1557, 1558, 1562, 1569, 1577, 1578, 1579, 
	1580, 1581, 1582, 1586, 1587, 1588, 1589, 1590, 
	1591, 1592, 1593, 1594, 1598, 1606, 1609, 1611, 
	1612, 1613, 1614, 1615, 1626, 1629, 1631, 1642, 
	1654, 1666, 1670, 1670, 1671, 1680, 1683, 1685, 
	1698, 1702, 1702, 1703, 1712, 1715, 1717, 1718, 
	1719, 1720, 1721, 1722, 1734, 1737, 1739, 1750, 
	1763, 1764, 1773, 1776, 1778, 1779, 1780, 1781, 
	1782, 1786, 1786, 1787, 1799, 1802, 1804, 1815, 
	1828, 1829, 1841, 1844, 1846, 1847, 1848, 1851, 
	1853, 1856, 1857, 1858, 1859, 1860, 1863, 1864, 
	1865
};

static const char _eo_tokenizer_trans_keys[] = {
	10, 42, 10, 42, 47, 10, 115, 116, 
	114, 97, 99, 116, 9, 13, 32, 9, 
	13, 32, 65, 90, 97, 122, 9, 10, 
	13, 32, 40, 95, 123, 0, 31, 48, 
	57, 65, 90, 97, 122, 10, 123, 0, 
	32, 9, 10, 13, 32, 40, 123, 0, 
	31, 41, 65, 90, 97, 122, 10, 123, 
	0, 32, 10, 41, 44, 95, 0, 32, 
	48, 57, 65, 90, 97, 122, 10, 41, 
	44, 0, 32, 10, 0, 32, 65, 90, 
	97, 122, 97, 115, 115, 9, 13, 32, 
	116, 101, 114, 102, 97, 99, 101, 9, 
	13, 32, 120, 105, 110, 9, 13, 32, 
	10, 42, 64, 10, 42, 10, 42, 47, 
	10, 42, 95, 0, 32, 48, 57, 64, 
	90, 97, 122, 10, 42, 10, 42, 47, 
	10, 9, 13, 32, 58, 9, 13, 32, 
	65, 90, 97, 122, 9, 13, 32, 65, 
	90, 97, 122, 9, 13, 32, 44, 59, 
	95, 48, 57, 65, 90, 97, 122, 9, 
	13, 32, 58, 95, 48, 57, 65, 90, 
	97, 122, 9, 13, 32, 58, 95, 103, 
	48, 57, 65, 90, 97, 122, 9, 13, 
	32, 58, 95, 97, 48, 57, 65, 90, 
	98, 122, 9, 13, 32, 58, 95, 99, 
	48, 57, 65, 90, 97, 122, 9, 13, 
	32, 58, 95, 121, 48, 57, 65, 90, 
	97, 122, 9, 13, 32, 58, 95, 48, 
	57, 65, 90, 97, 122, 9, 13, 32, 
	58, 65, 90, 97, 122, 59, 95, 48, 
	57, 65, 90, 97, 122, 9, 13, 32, 
	58, 95, 116, 48, 57, 65, 90, 97, 
	122, 9, 13, 32, 58, 95, 117, 48, 
	57, 65, 90, 97, 122, 9, 13, 32, 
	58, 95, 114, 48, 57, 65, 90, 97, 
	122, 9, 13, 32, 58, 95, 110, 48, 
	57, 65, 90, 97, 122, 9, 13, 32, 
	58, 95, 48, 57, 65, 90, 97, 122, 
	9, 13, 32, 58, 60, 62, 95, 40, 
	42, 45, 46, 48, 57, 64, 90, 97, 
	122, 9, 13, 32, 58, 60, 62, 95, 
	40, 42, 45, 46, 48, 57, 64, 90, 
	97, 122, 9, 13, 32, 58, 59, 62, 
	95, 40, 42, 45, 46, 48, 60, 64, 
	90, 97, 122, 9, 13, 32, 59, 60, 
	62, 95, 40, 42, 45, 46, 48, 57, 
	64, 90, 97, 122, 9, 13, 32, 47, 
	42, 64, 10, 95, 0, 32, 48, 57, 
	64, 90, 97, 122, 10, 42, 10, 42, 
	47, 9, 13, 32, 60, 62, 95, 40, 
	42, 45, 46, 48, 57, 64, 90, 97, 
	122, 10, 42, 10, 42, 47, 10, 9, 
	13, 32, 42, 59, 60, 62, 95, 48, 
	57, 64, 90, 97, 122, 9, 13, 32, 
	47, 42, 64, 10, 95, 0, 32, 48, 
	57, 64, 90, 97, 122, 10, 42, 10, 
	42, 47, 10, 42, 10, 42, 47, 10, 
	116, 10, 123, 0, 32, 121, 115, 10, 
	123, 0, 32, 116, 10, 123, 0, 32, 
	108, 117, 101, 115, 10, 123, 0, 32, 
	10, 42, 10, 42, 47, 10, 10, 123, 
	0, 32, 10, 95, 123, 0, 32, 48, 
	57, 65, 90, 97, 122, 10, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 111, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 116, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 101, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 99, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 116, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 101, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 100, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 9, 13, 32, 95, 48, 
	57, 65, 90, 97, 122, 9, 13, 32, 
	112, 48, 57, 65, 90, 97, 122, 10, 
	95, 114, 117, 123, 0, 32, 48, 57, 
	65, 90, 97, 122, 10, 95, 111, 123, 
	0, 32, 48, 57, 65, 90, 97, 122, 
	10, 95, 116, 123, 0, 32, 48, 57, 
	65, 90, 97, 122, 10, 95, 101, 123, 
	0, 32, 48, 57, 65, 90, 97, 122, 
	10, 95, 99, 123, 0, 32, 48, 57, 
	65, 90, 97, 122, 10, 95, 116, 123, 
	0, 32, 48, 57, 65, 90, 97, 122, 
	10, 95, 101, 123, 0, 32, 48, 57, 
	65, 90, 97, 122, 10, 95, 100, 123, 
	0, 32, 48, 57, 65, 90, 97, 122, 
	95, 48, 57, 65, 90, 97, 122, 10, 
	95, 98, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 108, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 105, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 99, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 98, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 108, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 105, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 99, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	42, 64, 10, 42, 10, 42, 47, 10, 
	42, 95, 0, 32, 48, 57, 64, 90, 
	97, 122, 10, 42, 10, 42, 47, 10, 
	110, 115, 116, 59, 103, 97, 99, 121, 
	9, 13, 32, 9, 13, 32, 65, 90, 
	97, 122, 59, 95, 48, 57, 65, 90, 
	97, 122, 114, 97, 109, 115, 10, 123, 
	0, 32, 116, 117, 114, 110, 9, 13, 
	32, 9, 13, 32, 60, 62, 95, 40, 
	42, 45, 46, 48, 57, 64, 90, 97, 
	122, 9, 13, 32, 60, 62, 95, 40, 
	42, 45, 46, 48, 57, 64, 90, 97, 
	122, 9, 13, 32, 59, 60, 62, 95, 
	40, 42, 45, 46, 48, 57, 64, 90, 
	97, 122, 9, 13, 32, 59, 60, 62, 
	95, 40, 42, 45, 46, 48, 57, 64, 
	90, 97, 122, 9, 13, 32, 47, 42, 
	64, 10, 95, 0, 32, 48, 57, 64, 
	90, 97, 122, 10, 42, 10, 42, 47, 
	9, 13, 32, 60, 62, 95, 40, 42, 
	45, 46, 48, 57, 64, 90, 97, 122, 
	10, 42, 10, 42, 47, 10, 10, 123, 
	0, 32, 10, 95, 123, 0, 32, 48, 
	57, 65, 90, 97, 122, 10, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 111, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 116, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 101, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 99, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 116, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 101, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 100, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 9, 13, 32, 95, 48, 
	57, 65, 90, 97, 122, 9, 13, 32, 
	112, 48, 57, 65, 90, 97, 122, 10, 
	95, 114, 117, 123, 0, 32, 48, 57, 
	65, 90, 97, 122, 10, 95, 111, 123, 
	0, 32, 48, 57, 65, 90, 97, 122, 
	10, 95, 116, 123, 0, 32, 48, 57, 
	65, 90, 97, 122, 10, 95, 101, 123, 
	0, 32, 48, 57, 65, 90, 97, 122, 
	10, 95, 99, 123, 0, 32, 48, 57, 
	65, 90, 97, 122, 10, 95, 116, 123, 
	0, 32, 48, 57, 65, 90, 97, 122, 
	10, 95, 101, 123, 0, 32, 48, 57, 
	65, 90, 97, 122, 10, 95, 100, 123, 
	0, 32, 48, 57, 65, 90, 97, 122, 
	95, 48, 57, 65, 90, 97, 122, 10, 
	95, 98, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 108, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 105, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 99, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 98, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 108, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 105, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 10, 95, 99, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	42, 64, 10, 42, 10, 42, 47, 10, 
	42, 95, 0, 32, 48, 57, 64, 90, 
	97, 122, 10, 42, 10, 42, 47, 10, 
	110, 115, 116, 114, 117, 99, 116, 111, 
	114, 115, 10, 123, 0, 32, 116, 97, 
	10, 58, 0, 32, 10, 0, 32, 65, 
	90, 97, 122, 59, 95, 48, 57, 65, 
	90, 97, 122, 95, 112, 114, 101, 102, 
	105, 120, 10, 58, 0, 32, 10, 0, 
	32, 65, 90, 97, 122, 59, 95, 48, 
	57, 65, 90, 97, 122, 101, 110, 116, 
	115, 10, 123, 0, 32, 10, 125, 0, 
	32, 65, 90, 97, 122, 9, 10, 13, 
	32, 40, 44, 59, 95, 0, 31, 48, 
	57, 65, 90, 97, 122, 10, 59, 0, 
	32, 9, 10, 13, 32, 47, 125, 0, 
	31, 65, 90, 97, 122, 42, 64, 10, 
	95, 0, 32, 48, 57, 64, 90, 97, 
	122, 10, 42, 10, 42, 47, 10, 125, 
	0, 32, 65, 90, 97, 122, 9, 10, 
	13, 32, 40, 59, 0, 31, 9, 13, 
	32, 95, 65, 90, 97, 122, 9, 13, 
	32, 41, 42, 95, 48, 57, 65, 90, 
	97, 122, 10, 59, 0, 32, 112, 108, 
	101, 109, 101, 110, 116, 115, 10, 123, 
	0, 32, 10, 125, 0, 32, 65, 90, 
	97, 122, 10, 58, 59, 95, 0, 32, 
	48, 57, 65, 90, 97, 122, 10, 59, 
	0, 32, 58, 103, 97, 99, 121, 95, 
	112, 114, 101, 102, 105, 120, 10, 58, 
	0, 32, 10, 0, 32, 65, 90, 97, 
	122, 59, 95, 48, 57, 65, 90, 97, 
	122, 116, 104, 111, 100, 115, 10, 123, 
	0, 32, 111, 112, 101, 114, 116, 105, 
	101, 115, 10, 123, 0, 32, 10, 47, 
	97, 99, 105, 109, 0, 32, 10, 0, 
	32, 42, 47, 98, 108, 110, 105, 10, 
	47, 108, 114, 125, 0, 32, 65, 90, 
	97, 122, 10, 0, 32, 42, 47, 9, 
	13, 32, 58, 95, 48, 57, 65, 90, 
	97, 122, 9, 13, 32, 58, 95, 101, 
	48, 57, 65, 90, 97, 122, 9, 13, 
	32, 58, 95, 101, 48, 57, 65, 90, 
	97, 122, 9, 13, 32, 47, 59, 10, 
	47, 125, 0, 32, 64, 90, 97, 122, 
	10, 0, 32, 42, 47, 9, 13, 32, 
	42, 60, 62, 95, 48, 57, 64, 90, 
	97, 122, 9, 13, 32, 47, 59, 10, 
	47, 103, 107, 115, 118, 125, 0, 32, 
	10, 0, 32, 42, 47, 101, 101, 101, 
	97, 59, 10, 47, 112, 125, 0, 32, 
	48, 57, 65, 90, 97, 122, 10, 0, 
	32, 42, 47, 10, 95, 123, 0, 32, 
	48, 57, 65, 90, 97, 122, 10, 95, 
	114, 117, 123, 0, 32, 48, 57, 65, 
	90, 97, 122, 59, 10, 47, 99, 108, 
	112, 114, 125, 0, 32, 10, 0, 32, 
	42, 47, 111, 101, 97, 101, 9, 13, 
	32, 47, 59, 10, 47, 112, 125, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	0, 32, 42, 47, 10, 95, 123, 0, 
	32, 48, 57, 65, 90, 97, 122, 10, 
	95, 114, 117, 123, 0, 32, 48, 57, 
	65, 90, 97, 122, 59, 10, 47, 99, 
	100, 101, 105, 108, 109, 112, 125, 0, 
	32, 10, 0, 32, 42, 47, 111, 97, 
	10, 0, 32, 111, 118, 10, 0, 32, 
	59, 109, 59, 101, 10, 0, 32, 101, 
	114, 59, 0
};

static const char _eo_tokenizer_single_lengths[] = {
	2, 3, 1, 1, 1, 1, 1, 1, 
	1, 3, 3, 7, 2, 6, 1, 2, 
	4, 3, 1, 1, 1, 1, 3, 1, 
	1, 1, 1, 1, 1, 1, 3, 1, 
	1, 1, 3, 3, 2, 3, 3, 2, 
	3, 1, 4, 3, 3, 6, 5, 6, 
	6, 6, 6, 5, 4, 2, 6, 6, 
	6, 6, 5, 7, 7, 7, 7, 4, 
	1, 1, 2, 2, 3, 6, 2, 3, 
	1, 8, 4, 1, 1, 2, 2, 3, 
	2, 3, 1, 1, 2, 1, 1, 2, 
	1, 2, 1, 1, 1, 1, 2, 2, 
	3, 1, 2, 3, 2, 4, 4, 4, 
	4, 4, 4, 4, 4, 4, 5, 4, 
	4, 4, 4, 4, 4, 4, 1, 4, 
	4, 4, 4, 4, 4, 4, 4, 3, 
	2, 3, 3, 2, 3, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 3, 3, 
	2, 1, 1, 1, 1, 2, 1, 1, 
	1, 1, 3, 6, 6, 7, 7, 4, 
	1, 1, 2, 2, 3, 6, 2, 3, 
	1, 2, 3, 2, 4, 4, 4, 4, 
	4, 4, 4, 4, 4, 5, 4, 4, 
	4, 4, 4, 4, 4, 1, 4, 4, 
	4, 4, 4, 4, 4, 4, 3, 2, 
	3, 3, 2, 3, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 2, 
	1, 1, 2, 1, 2, 1, 1, 1, 
	1, 1, 1, 1, 2, 1, 2, 1, 
	1, 1, 1, 2, 2, 8, 2, 6, 
	1, 1, 2, 2, 3, 2, 6, 4, 
	6, 2, 1, 1, 1, 1, 1, 1, 
	1, 1, 2, 2, 4, 2, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 2, 1, 2, 1, 1, 1, 
	1, 1, 2, 1, 1, 1, 1, 1, 
	1, 1, 1, 2, 6, 1, 2, 1, 
	1, 1, 1, 5, 1, 2, 5, 6, 
	6, 4, 0, 1, 3, 1, 2, 7, 
	4, 0, 1, 7, 1, 2, 1, 1, 
	1, 1, 1, 4, 1, 2, 3, 5, 
	1, 7, 1, 2, 1, 1, 1, 1, 
	4, 0, 1, 4, 1, 2, 3, 5, 
	1, 10, 1, 2, 1, 1, 1, 2, 
	1, 1, 1, 1, 1, 1, 1, 1, 
	1
};

static const char _eo_tokenizer_range_lengths[] = {
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 2, 4, 1, 1, 2, 1, 
	4, 1, 3, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 4, 0, 
	0, 0, 0, 2, 2, 3, 3, 3, 
	3, 3, 3, 3, 2, 3, 3, 3, 
	3, 3, 3, 5, 5, 5, 5, 0, 
	0, 0, 4, 0, 0, 5, 0, 0, 
	0, 3, 0, 0, 0, 4, 0, 0, 
	0, 0, 0, 0, 1, 0, 0, 1, 
	0, 1, 0, 0, 0, 0, 1, 0, 
	0, 0, 1, 4, 4, 4, 4, 4, 
	4, 4, 4, 4, 3, 3, 4, 4, 
	4, 4, 4, 4, 4, 4, 3, 4, 
	4, 4, 4, 4, 4, 4, 4, 0, 
	0, 0, 4, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 2, 
	3, 0, 0, 0, 0, 1, 0, 0, 
	0, 0, 0, 5, 5, 5, 5, 0, 
	0, 0, 4, 0, 0, 5, 0, 0, 
	0, 1, 4, 4, 4, 4, 4, 4, 
	4, 4, 4, 3, 3, 4, 4, 4, 
	4, 4, 4, 4, 4, 3, 4, 4, 
	4, 4, 4, 4, 4, 4, 0, 0, 
	0, 4, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 1, 
	0, 0, 1, 3, 3, 0, 0, 0, 
	0, 0, 0, 0, 1, 3, 3, 0, 
	0, 0, 0, 1, 3, 4, 1, 3, 
	0, 0, 4, 0, 0, 3, 1, 2, 
	3, 1, 0, 0, 0, 0, 0, 0, 
	0, 0, 1, 3, 4, 1, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 1, 3, 3, 0, 0, 0, 
	0, 0, 1, 0, 0, 0, 0, 0, 
	0, 0, 0, 1, 1, 1, 0, 0, 
	0, 0, 0, 3, 1, 0, 3, 3, 
	3, 0, 0, 0, 3, 1, 0, 3, 
	0, 0, 0, 1, 1, 0, 0, 0, 
	0, 0, 0, 4, 1, 0, 4, 4, 
	0, 1, 1, 0, 0, 0, 0, 0, 
	0, 0, 0, 4, 1, 0, 4, 4, 
	0, 1, 1, 0, 0, 0, 1, 0, 
	1, 0, 0, 0, 0, 1, 0, 0, 
	0
};

static const short _eo_tokenizer_index_offsets[] = {
	0, 3, 7, 9, 11, 13, 15, 17, 
	19, 21, 25, 31, 43, 47, 55, 59, 
	63, 72, 77, 82, 84, 86, 88, 92, 
	94, 96, 98, 100, 102, 104, 106, 110, 
	112, 114, 116, 120, 124, 127, 131, 139, 
	142, 146, 148, 153, 159, 165, 175, 184, 
	194, 204, 214, 224, 233, 240, 246, 256, 
	266, 276, 286, 295, 308, 321, 334, 347, 
	352, 354, 356, 363, 366, 370, 382, 385, 
	389, 391, 403, 408, 410, 412, 419, 422, 
	426, 429, 433, 435, 437, 441, 443, 445, 
	449, 451, 455, 457, 459, 461, 463, 467, 
	470, 474, 476, 480, 488, 495, 504, 513, 
	522, 531, 540, 549, 558, 566, 574, 584, 
	593, 602, 611, 620, 629, 638, 647, 652, 
	661, 670, 679, 688, 697, 706, 715, 724, 
	728, 731, 735, 743, 746, 750, 752, 754, 
	756, 758, 760, 762, 764, 766, 768, 772, 
	778, 784, 786, 788, 790, 792, 796, 798, 
	800, 802, 804, 808, 820, 832, 845, 858, 
	863, 865, 867, 874, 877, 881, 893, 896, 
	900, 902, 906, 914, 921, 930, 939, 948, 
	957, 966, 975, 984, 992, 1000, 1010, 1019, 
	1028, 1037, 1046, 1055, 1064, 1073, 1078, 1087, 
	1096, 1105, 1114, 1123, 1132, 1141, 1150, 1154, 
	1157, 1161, 1169, 1172, 1176, 1178, 1180, 1182, 
	1184, 1186, 1188, 1190, 1192, 1194, 1196, 1198, 
	1202, 1204, 1206, 1210, 1215, 1221, 1223, 1225, 
	1227, 1229, 1231, 1233, 1235, 1239, 1244, 1250, 
	1252, 1254, 1256, 1258, 1262, 1268, 1281, 1285, 
	1295, 1297, 1299, 1306, 1309, 1313, 1319, 1327, 
	1334, 1344, 1348, 1350, 1352, 1354, 1356, 1358, 
	1360, 1362, 1364, 1368, 1374, 1383, 1387, 1389, 
	1391, 1393, 1395, 1397, 1399, 1401, 1403, 1405, 
	1407, 1409, 1411, 1415, 1420, 1426, 1428, 1430, 
	1432, 1434, 1436, 1440, 1442, 1444, 1446, 1448, 
	1450, 1452, 1454, 1456, 1460, 1468, 1471, 1474, 
	1476, 1478, 1480, 1482, 1491, 1494, 1497, 1506, 
	1516, 1526, 1531, 1532, 1534, 1541, 1544, 1547, 
	1558, 1563, 1564, 1566, 1575, 1578, 1581, 1583, 
	1585, 1587, 1589, 1591, 1600, 1603, 1606, 1614, 
	1624, 1626, 1635, 1638, 1641, 1643, 1645, 1647, 
	1649, 1654, 1655, 1657, 1666, 1669, 1672, 1680, 
	1690, 1692, 1704, 1707, 1710, 1712, 1714, 1717, 
	1720, 1723, 1725, 1727, 1729, 1731, 1734, 1736, 
	1738
};

static const short _eo_tokenizer_indicies[] = {
	2, 3, 1, 2, 3, 4, 1, 6, 
	5, 7, 0, 8, 0, 9, 0, 10, 
	0, 11, 0, 12, 0, 13, 13, 13, 
	0, 14, 14, 14, 15, 15, 0, 17, 
	18, 17, 17, 19, 20, 21, 16, 20, 
	20, 20, 0, 23, 24, 22, 0, 25, 
	23, 25, 25, 26, 24, 22, 0, 27, 
	28, 28, 0, 30, 31, 29, 0, 33, 
	34, 35, 36, 32, 36, 36, 36, 0, 
	38, 27, 39, 37, 0, 40, 39, 28, 
	28, 0, 41, 0, 42, 0, 43, 0, 
	44, 44, 44, 0, 45, 0, 46, 0, 
	47, 0, 48, 0, 49, 0, 50, 0, 
	51, 0, 52, 52, 52, 0, 53, 0, 
	54, 0, 55, 0, 56, 56, 56, 0, 
	59, 60, 61, 58, 59, 60, 58, 59, 
	60, 62, 58, 63, 60, 64, 61, 64, 
	64, 64, 58, 66, 67, 65, 66, 67, 
	68, 65, 70, 69, 71, 71, 71, 72, 
	57, 73, 73, 73, 74, 74, 57, 75, 
	75, 75, 76, 76, 57, 77, 77, 77, 
	77, 78, 77, 77, 77, 77, 57, 71, 
	71, 71, 72, 79, 79, 79, 79, 57, 
	71, 71, 71, 72, 79, 80, 79, 79, 
	79, 57, 71, 71, 71, 72, 79, 81, 
	79, 79, 79, 57, 71, 71, 71, 72, 
	79, 82, 79, 79, 79, 57, 71, 71, 
	71, 72, 79, 83, 79, 79, 79, 57, 
	84, 84, 84, 72, 79, 79, 79, 79, 
	57, 84, 84, 84, 72, 85, 85, 57, 
	87, 86, 86, 86, 86, 57, 71, 71, 
	71, 72, 79, 88, 79, 79, 79, 57, 
	71, 71, 71, 72, 79, 89, 79, 79, 
	79, 57, 71, 71, 71, 72, 79, 90, 
	79, 79, 79, 57, 71, 71, 71, 72, 
	79, 91, 79, 79, 79, 57, 92, 92, 
	92, 72, 79, 79, 79, 79, 57, 93, 
	93, 93, 72, 94, 94, 94, 94, 94, 
	94, 94, 94, 57, 95, 95, 95, 72, 
	96, 96, 96, 96, 96, 96, 96, 96, 
	57, 95, 95, 95, 72, 97, 96, 96, 
	96, 96, 96, 96, 96, 57, 98, 98, 
	98, 97, 98, 98, 98, 98, 98, 98, 
	98, 98, 57, 100, 100, 100, 101, 99, 
	102, 99, 103, 99, 104, 105, 103, 105, 
	105, 105, 99, 107, 108, 106, 107, 108, 
	109, 106, 98, 98, 98, 98, 98, 98, 
	98, 98, 98, 98, 98, 57, 112, 113, 
	111, 112, 113, 114, 111, 116, 115, 117, 
	117, 117, 117, 118, 117, 117, 117, 117, 
	117, 117, 110, 120, 120, 120, 121, 119, 
	122, 119, 123, 119, 124, 125, 123, 125, 
	125, 125, 119, 127, 128, 126, 127, 128, 
	129, 126, 132, 133, 131, 132, 133, 134, 
	131, 136, 135, 137, 130, 138, 139, 137, 
	130, 140, 130, 141, 130, 142, 143, 141, 
	130, 144, 130, 145, 146, 144, 130, 147, 
	130, 148, 130, 149, 130, 150, 130, 151, 
	152, 150, 130, 155, 156, 154, 155, 156, 
	157, 154, 159, 158, 161, 162, 160, 153, 
	164, 166, 167, 163, 165, 165, 165, 153, 
	164, 167, 163, 165, 165, 165, 153, 164, 
	166, 168, 167, 163, 165, 165, 165, 153, 
	164, 166, 169, 167, 163, 165, 165, 165, 
	153, 164, 166, 170, 167, 163, 165, 165, 
	165, 153, 164, 166, 171, 167, 163, 165, 
	165, 165, 153, 164, 166, 172, 167, 163, 
	165, 165, 165, 153, 164, 166, 173, 167, 
	163, 165, 165, 165, 153, 164, 166, 174, 
	167, 163, 165, 165, 165, 153, 175, 175, 
	175, 166, 165, 165, 165, 153, 175, 175, 
	175, 177, 176, 176, 176, 153, 164, 166, 
	178, 179, 167, 163, 165, 165, 165, 153, 
	164, 166, 180, 167, 163, 165, 165, 165, 
	153, 164, 166, 181, 167, 163, 165, 165, 
	165, 153, 164, 166, 182, 167, 163, 165, 
	165, 165, 153, 164, 166, 183, 167, 163, 
	165, 165, 165, 153, 164, 166, 184, 167, 
	163, 165, 165, 165, 153, 164, 166, 185, 
	167, 163, 165, 165, 165, 153, 164, 166, 
	186, 167, 163, 165, 165, 165, 153, 166, 
	165, 165, 165, 153, 164, 166, 187, 167, 
	163, 165, 165, 165, 153, 164, 166, 188, 
	167, 163, 165, 165, 165, 153, 164, 166, 
	189, 167, 163, 165, 165, 165, 153, 164, 
	166, 186, 167, 163, 165, 165, 165, 153, 
	164, 166, 190, 167, 163, 165, 165, 165, 
	153, 164, 166, 191, 167, 163, 165, 165, 
	165, 153, 164, 166, 192, 167, 163, 165, 
	165, 165, 153, 164, 166, 174, 167, 163, 
	165, 165, 165, 153, 195, 196, 197, 194, 
	195, 196, 194, 195, 196, 198, 194, 199, 
	196, 200, 197, 200, 200, 200, 194, 202, 
	203, 201, 202, 203, 204, 201, 206, 205, 
	207, 193, 208, 193, 209, 193, 210, 193, 
	211, 193, 212, 193, 213, 193, 214, 193, 
	215, 215, 215, 193, 215, 215, 215, 216, 
	216, 193, 218, 217, 217, 217, 217, 193, 
	219, 193, 220, 193, 221, 193, 222, 193, 
	223, 224, 222, 193, 225, 193, 226, 193, 
	227, 193, 228, 193, 229, 229, 229, 193, 
	230, 230, 230, 231, 231, 231, 231, 231, 
	231, 231, 231, 193, 232, 232, 232, 233, 
	233, 233, 233, 233, 233, 233, 233, 193, 
	232, 232, 232, 234, 233, 233, 233, 233, 
	233, 233, 233, 233, 193, 235, 235, 235, 
	234, 235, 235, 235, 235, 235, 235, 235, 
	235, 193, 237, 237, 237, 238, 236, 239, 
	236, 240, 236, 241, 242, 240, 242, 242, 
	242, 236, 244, 245, 243, 244, 245, 246, 
	243, 235, 235, 235, 235, 235, 235, 235, 
	235, 235, 235, 235, 193, 249, 250, 248, 
	249, 250, 251, 248, 253, 252, 255, 256, 
	254, 247, 258, 260, 261, 257, 259, 259, 
	259, 247, 258, 261, 257, 259, 259, 259, 
	247, 258, 260, 262, 261, 257, 259, 259, 
	259, 247, 258, 260, 263, 261, 257, 259, 
	259, 259, 247, 258, 260, 264, 261, 257, 
	259, 259, 259, 247, 258, 260, 265, 261, 
	257, 259, 259, 259, 247, 258, 260, 266, 
	261, 257, 259, 259, 259, 247, 258, 260, 
	267, 261, 257, 259, 259, 259, 247, 258, 
	260, 268, 261, 257, 259, 259, 259, 247, 
	269, 269, 269, 260, 259, 259, 259, 247, 
	269, 269, 269, 271, 270, 270, 270, 247, 
	258, 260, 272, 273, 261, 257, 259, 259, 
	259, 247, 258, 260, 274, 261, 257, 259, 
	259, 259, 247, 258, 260, 275, 261, 257, 
	259, 259, 259, 247, 258, 260, 276, 261, 
	257, 259, 259, 259, 247, 258, 260, 277, 
	261, 257, 259, 259, 259, 247, 258, 260, 
	278, 261, 257, 259, 259, 259, 247, 258, 
	260, 279, 261, 257, 259, 259, 259, 247, 
	258, 260, 280, 261, 257, 259, 259, 259, 
	247, 260, 259, 259, 259, 247, 258, 260, 
	281, 261, 257, 259, 259, 259, 247, 258, 
	260, 282, 261, 257, 259, 259, 259, 247, 
	258, 260, 283, 261, 257, 259, 259, 259, 
	247, 258, 260, 280, 261, 257, 259, 259, 
	259, 247, 258, 260, 284, 261, 257, 259, 
	259, 259, 247, 258, 260, 285, 261, 257, 
	259, 259, 259, 247, 258, 260, 286, 261, 
	257, 259, 259, 259, 247, 258, 260, 268, 
	261, 257, 259, 259, 259, 247, 289, 290, 
	291, 288, 289, 290, 288, 289, 290, 292, 
	288, 293, 290, 294, 291, 294, 294, 294, 
	288, 296, 297, 295, 296, 297, 298, 295, 
	300, 299, 301, 287, 302, 287, 303, 287, 
	304, 287, 305, 287, 306, 287, 307, 287, 
	308, 287, 309, 287, 310, 287, 311, 312, 
	310, 287, 313, 287, 314, 287, 315, 316, 
	314, 287, 317, 316, 318, 318, 287, 320, 
	319, 319, 319, 319, 287, 321, 287, 322, 
	287, 323, 287, 324, 287, 325, 287, 326, 
	287, 327, 287, 328, 329, 327, 287, 330, 
	329, 331, 331, 287, 333, 332, 332, 332, 
	332, 287, 334, 287, 335, 287, 336, 287, 
	337, 287, 338, 339, 337, 287, 340, 342, 
	339, 341, 341, 287, 344, 345, 344, 344, 
	346, 347, 348, 347, 343, 347, 347, 347, 
	287, 350, 351, 349, 287, 351, 340, 351, 
	351, 352, 342, 339, 341, 341, 287, 353, 
	287, 354, 287, 355, 356, 354, 356, 356, 
	356, 287, 358, 359, 357, 358, 359, 360, 
	357, 362, 364, 361, 363, 363, 287, 365, 
	350, 365, 365, 366, 351, 349, 287, 366, 
	366, 366, 367, 367, 367, 287, 368, 368, 
	368, 369, 368, 368, 368, 368, 368, 287, 
	371, 372, 370, 287, 373, 287, 374, 287, 
	375, 287, 376, 287, 377, 287, 378, 287, 
	379, 287, 380, 287, 381, 382, 380, 287, 
	383, 385, 382, 384, 384, 287, 387, 389, 
	390, 388, 386, 388, 388, 388, 287, 392, 
	382, 391, 287, 388, 287, 393, 287, 394, 
	287, 395, 287, 396, 287, 397, 287, 398, 
	287, 399, 287, 400, 287, 401, 287, 402, 
	287, 403, 287, 404, 405, 403, 287, 406, 
	405, 407, 407, 287, 409, 408, 408, 408, 
	408, 287, 410, 287, 411, 287, 412, 287, 
	413, 287, 414, 287, 415, 416, 414, 287, 
	417, 287, 418, 287, 419, 287, 420, 287, 
	421, 287, 422, 287, 423, 287, 424, 287, 
	425, 426, 424, 287, 429, 430, 431, 432, 
	433, 434, 428, 427, 429, 428, 435, 1, 
	5, 436, 437, 436, 438, 436, 439, 436, 
	440, 436, 443, 444, 446, 447, 448, 442, 
	445, 445, 441, 443, 442, 449, 451, 69, 
	450, 71, 71, 71, 72, 79, 79, 79, 
	79, 450, 71, 71, 71, 72, 79, 452, 
	79, 79, 79, 450, 71, 71, 71, 72, 
	79, 453, 79, 79, 79, 450, 100, 100, 
	100, 101, 454, 455, 457, 456, 460, 461, 
	463, 459, 462, 462, 458, 460, 459, 464, 
	111, 115, 465, 117, 117, 117, 117, 117, 
	117, 117, 117, 117, 117, 465, 120, 120, 
	120, 121, 466, 467, 469, 468, 472, 473, 
	474, 475, 476, 477, 478, 471, 470, 472, 
	471, 479, 131, 135, 480, 481, 480, 482, 
	480, 483, 480, 484, 480, 486, 485, 489, 
	490, 492, 493, 488, 491, 491, 491, 487, 
	489, 488, 494, 154, 158, 495, 164, 166, 
	167, 163, 165, 165, 165, 495, 164, 166, 
	496, 497, 167, 163, 165, 165, 165, 495, 
	499, 498, 502, 503, 504, 505, 506, 507, 
	508, 501, 500, 502, 501, 509, 511, 205, 
	510, 512, 510, 513, 510, 514, 510, 515, 
	510, 237, 237, 237, 238, 516, 517, 519, 
	518, 522, 523, 525, 526, 521, 524, 524, 
	524, 520, 522, 521, 527, 248, 252, 528, 
	258, 260, 261, 257, 259, 259, 259, 528, 
	258, 260, 529, 530, 261, 257, 259, 259, 
	259, 528, 532, 531, 535, 536, 537, 538, 
	539, 540, 541, 542, 543, 544, 534, 533, 
	535, 534, 545, 547, 299, 546, 548, 546, 
	549, 546, 552, 551, 550, 553, 554, 546, 
	557, 556, 555, 559, 558, 560, 546, 562, 
	561, 563, 546, 566, 565, 564, 567, 546, 
	568, 546, 570, 569, 0
};

static const short _eo_tokenizer_trans_targs[] = {
	292, 0, 0, 1, 292, 2, 292, 4, 
	5, 6, 7, 8, 9, 10, 10, 11, 
	12, 13, 12, 14, 11, 292, 12, 12, 
	292, 13, 14, 15, 16, 12, 12, 292, 
	17, 17, 15, 18, 16, 17, 17, 18, 
	18, 20, 21, 22, 10, 24, 25, 26, 
	27, 28, 29, 30, 10, 32, 33, 34, 
	10, 299, 36, 36, 37, 38, 299, 38, 
	39, 39, 39, 40, 299, 41, 299, 42, 
	43, 44, 45, 44, 45, 45, 299, 46, 
	48, 49, 50, 51, 52, 53, 53, 299, 
	55, 56, 57, 58, 59, 60, 69, 61, 
	62, 305, 62, 299, 63, 64, 65, 66, 
	66, 67, 67, 67, 68, 306, 308, 70, 
	70, 71, 308, 72, 308, 73, 312, 308, 
	74, 75, 76, 77, 77, 78, 78, 78, 
	79, 313, 315, 80, 80, 81, 315, 82, 
	315, 84, 84, 315, 86, 87, 87, 315, 
	89, 89, 315, 91, 92, 93, 94, 94, 
	315, 323, 95, 95, 96, 323, 97, 323, 
	98, 98, 323, 98, 98, 99, 100, 323, 
	102, 103, 104, 105, 106, 107, 108, 109, 
	99, 110, 111, 119, 112, 113, 114, 115, 
	116, 117, 118, 120, 121, 122, 124, 125, 
	126, 329, 128, 128, 129, 130, 329, 130, 
	131, 131, 131, 132, 329, 133, 329, 135, 
	136, 137, 329, 139, 140, 141, 142, 143, 
	144, 144, 329, 146, 147, 148, 149, 149, 
	329, 151, 152, 153, 154, 155, 156, 165, 
	157, 158, 336, 158, 329, 159, 160, 161, 
	162, 162, 163, 163, 163, 164, 337, 339, 
	166, 166, 167, 339, 168, 339, 169, 169, 
	339, 169, 169, 170, 171, 339, 173, 174, 
	175, 176, 177, 178, 179, 180, 170, 181, 
	182, 190, 183, 184, 185, 186, 187, 188, 
	189, 191, 192, 193, 195, 196, 197, 345, 
	199, 199, 200, 201, 345, 201, 202, 202, 
	202, 203, 345, 204, 345, 206, 207, 208, 
	209, 210, 211, 212, 213, 214, 215, 215, 
	345, 217, 218, 218, 219, 219, 220, 220, 
	350, 222, 223, 224, 225, 226, 227, 228, 
	228, 229, 229, 230, 230, 352, 232, 233, 
	234, 235, 235, 236, 236, 237, 353, 238, 
	246, 238, 247, 237, 239, 238, 238, 239, 
	240, 241, 242, 242, 243, 243, 243, 244, 
	245, 236, 236, 237, 353, 246, 247, 248, 
	248, 249, 238, 238, 239, 251, 252, 253, 
	254, 255, 256, 257, 258, 258, 259, 259, 
	260, 355, 261, 261, 260, 262, 259, 261, 
	261, 264, 265, 266, 267, 268, 269, 270, 
	271, 272, 273, 274, 274, 275, 275, 276, 
	276, 357, 278, 279, 280, 281, 282, 282, 
	345, 284, 285, 286, 287, 288, 289, 290, 
	291, 291, 345, 292, 293, 293, 294, 295, 
	296, 297, 298, 292, 292, 3, 19, 23, 
	31, 299, 300, 300, 301, 302, 303, 304, 
	307, 299, 299, 35, 47, 54, 299, 299, 
	299, 299, 308, 309, 309, 310, 311, 314, 
	308, 308, 308, 308, 308, 308, 315, 316, 
	316, 317, 318, 319, 320, 321, 322, 315, 
	315, 83, 85, 88, 90, 315, 315, 323, 
	324, 324, 325, 326, 327, 328, 323, 323, 
	101, 123, 323, 323, 329, 330, 330, 331, 
	332, 333, 334, 335, 338, 329, 329, 127, 
	134, 138, 145, 150, 329, 329, 329, 329, 
	339, 340, 340, 341, 342, 343, 344, 339, 
	339, 172, 194, 339, 339, 345, 346, 346, 
	347, 348, 349, 351, 354, 356, 358, 359, 
	360, 345, 345, 198, 205, 216, 345, 350, 
	350, 221, 231, 345, 352, 352, 345, 345, 
	250, 345, 345, 263, 345, 357, 357, 277, 
	283, 345, 345
};

static const short _eo_tokenizer_trans_actions[] = {
	205, 0, 1, 0, 195, 0, 228, 0, 
	0, 0, 0, 0, 0, 31, 0, 3, 
	37, 37, 291, 37, 0, 294, 0, 1, 
	197, 0, 0, 0, 3, 13, 270, 273, 
	11, 267, 11, 11, 0, 0, 1, 0, 
	1, 0, 0, 0, 29, 0, 0, 0, 
	0, 0, 0, 0, 35, 0, 0, 0, 
	33, 63, 0, 1, 0, 0, 47, 1, 
	3, 0, 1, 0, 45, 0, 207, 0, 
	0, 5, 237, 0, 3, 0, 240, 0, 
	0, 0, 0, 0, 0, 3, 0, 234, 
	0, 0, 0, 0, 0, 3, 3, 3, 
	3, 303, 0, 61, 0, 0, 0, 0, 
	1, 3, 0, 1, 0, 0, 81, 0, 
	1, 0, 65, 0, 210, 0, 306, 79, 
	0, 0, 0, 0, 1, 3, 0, 1, 
	0, 0, 103, 0, 1, 0, 83, 0, 
	213, 0, 1, 85, 0, 0, 1, 89, 
	0, 1, 87, 0, 0, 0, 0, 1, 
	91, 119, 0, 1, 0, 105, 0, 216, 
	0, 1, 107, 7, 246, 0, 0, 249, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	312, 312, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 141, 0, 1, 0, 0, 123, 1, 
	3, 0, 1, 0, 121, 0, 219, 0, 
	0, 0, 258, 0, 0, 0, 0, 0, 
	3, 0, 255, 0, 0, 0, 0, 1, 
	125, 0, 0, 0, 0, 0, 3, 3, 
	3, 3, 309, 0, 139, 0, 0, 0, 
	0, 1, 3, 0, 1, 0, 0, 157, 
	0, 1, 0, 143, 0, 222, 0, 1, 
	145, 9, 261, 0, 0, 264, 0, 0, 
	0, 0, 0, 0, 0, 0, 316, 316, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 193, 
	0, 1, 0, 0, 161, 1, 3, 0, 
	1, 0, 159, 0, 225, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 1, 
	167, 0, 0, 1, 0, 1, 3, 0, 
	25, 0, 0, 0, 0, 0, 0, 0, 
	1, 0, 1, 3, 0, 23, 0, 0, 
	0, 0, 1, 0, 1, 3, 0, 15, 
	15, 276, 15, 0, 15, 0, 1, 0, 
	0, 0, 0, 1, 3, 0, 1, 0, 
	0, 19, 282, 285, 19, 0, 0, 3, 
	0, 0, 17, 279, 17, 0, 0, 0, 
	0, 0, 0, 0, 0, 1, 0, 1, 
	3, 0, 27, 288, 0, 0, 27, 0, 
	1, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 1, 0, 1, 3, 
	0, 21, 0, 0, 0, 0, 0, 1, 
	171, 0, 0, 0, 0, 0, 0, 0, 
	0, 1, 169, 199, 0, 1, 297, 43, 
	43, 43, 43, 201, 203, 0, 0, 0, 
	0, 51, 0, 1, 297, 300, 300, 300, 
	0, 53, 59, 0, 0, 0, 55, 231, 
	57, 49, 69, 0, 1, 297, 300, 0, 
	71, 77, 73, 243, 75, 67, 95, 0, 
	1, 297, 43, 43, 43, 43, 0, 97, 
	101, 0, 0, 0, 0, 99, 93, 111, 
	0, 1, 297, 300, 300, 0, 113, 117, 
	0, 0, 115, 109, 129, 0, 1, 297, 
	43, 43, 43, 43, 0, 131, 137, 0, 
	0, 0, 0, 0, 133, 252, 135, 127, 
	149, 0, 1, 297, 300, 300, 0, 151, 
	155, 0, 0, 153, 147, 175, 0, 1, 
	297, 43, 43, 43, 43, 43, 43, 43, 
	0, 177, 191, 0, 0, 0, 183, 0, 
	1, 0, 0, 181, 0, 1, 187, 165, 
	0, 185, 163, 0, 179, 0, 1, 0, 
	0, 189, 173
};

static const short _eo_tokenizer_to_state_actions[] = {
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 39, 0, 0, 0, 
	0, 0, 0, 39, 0, 0, 0, 0, 
	0, 0, 0, 0, 39, 0, 0, 0, 
	0, 0, 0, 39, 0, 0, 0, 0, 
	0, 0, 0, 39, 0, 0, 0, 0, 
	0, 39, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 39, 0, 0, 0, 0, 
	0, 39, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0
};

static const short _eo_tokenizer_from_state_actions[] = {
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 41, 0, 0, 0, 
	0, 0, 0, 41, 0, 0, 0, 0, 
	0, 0, 0, 0, 41, 0, 0, 0, 
	0, 0, 0, 41, 0, 0, 0, 0, 
	0, 0, 0, 41, 0, 0, 0, 0, 
	0, 41, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 41, 0, 0, 0, 0, 
	0, 41, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0
};

static const short _eo_tokenizer_eof_trans[] = {
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 58, 58, 58, 58, 58, 
	58, 58, 58, 58, 58, 58, 58, 58, 
	58, 58, 58, 58, 58, 58, 58, 58, 
	58, 58, 58, 58, 58, 58, 58, 100, 
	100, 100, 100, 100, 100, 58, 111, 111, 
	111, 111, 120, 120, 120, 120, 120, 120, 
	131, 131, 131, 131, 131, 131, 131, 131, 
	131, 131, 131, 131, 131, 131, 131, 154, 
	154, 154, 154, 154, 154, 154, 154, 154, 
	154, 154, 154, 154, 154, 154, 154, 154, 
	154, 154, 154, 154, 154, 154, 154, 154, 
	154, 154, 154, 154, 154, 154, 154, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 194, 
	194, 194, 194, 194, 194, 194, 194, 237, 
	237, 237, 237, 237, 237, 194, 248, 248, 
	248, 248, 248, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 248, 248, 
	248, 248, 248, 248, 248, 248, 288, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 288, 288, 288, 288, 288, 288, 288, 
	288, 288, 288, 288, 0, 436, 437, 437, 
	437, 437, 437, 0, 450, 451, 451, 451, 
	451, 455, 456, 457, 0, 465, 466, 466, 
	467, 468, 469, 0, 480, 481, 481, 481, 
	481, 481, 486, 0, 495, 496, 496, 496, 
	499, 0, 510, 511, 511, 511, 511, 511, 
	517, 518, 519, 0, 528, 529, 529, 529, 
	532, 0, 546, 547, 547, 547, 551, 547, 
	556, 559, 547, 562, 547, 565, 547, 547, 
	570
};

static const int eo_tokenizer_start = 292;
static const int eo_tokenizer_first_final = 292;
static const int eo_tokenizer_error = -1;

static const int eo_tokenizer_en_tokenize_accessor = 299;
static const int eo_tokenizer_en_tokenize_params = 308;
static const int eo_tokenizer_en_tokenize_property = 315;
static const int eo_tokenizer_en_tokenize_properties = 323;
static const int eo_tokenizer_en_tokenize_method = 329;
static const int eo_tokenizer_en_tokenize_methods = 339;
static const int eo_tokenizer_en_tokenize_class = 345;
static const int eo_tokenizer_en_main = 292;


#line 968 "lib/eolian/eo_lexer.rl"


Eina_Bool
eo_tokenizer_walk(Eo_Tokenizer *toknz, const char *source)
{
   INF("tokenize %s...", source);
   toknz->source = eina_stringshare_add(source);

   FILE *stream;
   Eina_Bool ret = EINA_TRUE;

   int done = 0;
   int have = 0;
   int offset = 0;

   stream = fopen(toknz->source, "rb");
   if (!stream)
     {
        ERR("unable to read in %s", toknz->source);
        return EINA_FALSE;
     }

   
#line 1400 "lib/eolian/eo_lexer.c"
	{
	 toknz->cs = eo_tokenizer_start;
	 toknz->ts = 0;
	 toknz->te = 0;
	 toknz->act = 0;
	}

#line 991 "lib/eolian/eo_lexer.rl"

   while (!done)
     {
        int len;
        int space;

        toknz->p = toknz->buf + have;
        space = BUFSIZE - have;

        if (space == 0)
          {
             fclose(stream);
             ABORT(toknz, "out of buffer space");
          }

        len = fread(toknz->p, 1, space, stream);
        if (len == 0) break;
        toknz->pe = toknz->p + len;

        if (len < space)
          {
             toknz->eof = toknz->pe;
             done = 1;
          }

        
#line 1435 "lib/eolian/eo_lexer.c"
	{
	int _klen;
	unsigned int _trans;
	const char *_acts;
	unsigned int _nacts;
	const char *_keys;

	if ( ( toknz->p) == ( toknz->pe) )
		goto _test_eof;
_resume:
	_acts = _eo_tokenizer_actions + _eo_tokenizer_from_state_actions[ toknz->cs];
	_nacts = (unsigned int) *_acts++;
	while ( _nacts-- > 0 ) {
		switch ( *_acts++ ) {
	case 34:
#line 1 "NONE"
	{ toknz->ts = ( toknz->p);}
	break;
#line 1454 "lib/eolian/eo_lexer.c"
		}
	}

	_keys = _eo_tokenizer_trans_keys + _eo_tokenizer_key_offsets[ toknz->cs];
	_trans = _eo_tokenizer_index_offsets[ toknz->cs];

	_klen = _eo_tokenizer_single_lengths[ toknz->cs];
	if ( _klen > 0 ) {
		const char *_lower = _keys;
		const char *_mid;
		const char *_upper = _keys + _klen - 1;
		while (1) {
			if ( _upper < _lower )
				break;

			_mid = _lower + ((_upper-_lower) >> 1);
			if ( (*( toknz->p)) < *_mid )
				_upper = _mid - 1;
			else if ( (*( toknz->p)) > *_mid )
				_lower = _mid + 1;
			else {
				_trans += (unsigned int)(_mid - _keys);
				goto _match;
			}
		}
		_keys += _klen;
		_trans += _klen;
	}

	_klen = _eo_tokenizer_range_lengths[ toknz->cs];
	if ( _klen > 0 ) {
		const char *_lower = _keys;
		const char *_mid;
		const char *_upper = _keys + (_klen<<1) - 2;
		while (1) {
			if ( _upper < _lower )
				break;

			_mid = _lower + (((_upper-_lower) >> 1) & ~1);
			if ( (*( toknz->p)) < _mid[0] )
				_upper = _mid - 2;
			else if ( (*( toknz->p)) > _mid[1] )
				_lower = _mid + 2;
			else {
				_trans += (unsigned int)((_mid - _keys)>>1);
				goto _match;
			}
		}
		_trans += _klen;
	}

_match:
	_trans = _eo_tokenizer_indicies[_trans];
_eof_trans:
	 toknz->cs = _eo_tokenizer_trans_targs[_trans];

	if ( _eo_tokenizer_trans_actions[_trans] == 0 )
		goto _again;

	_acts = _eo_tokenizer_actions + _eo_tokenizer_trans_actions[_trans];
	_nacts = (unsigned int) *_acts++;
	while ( _nacts-- > 0 )
	{
		switch ( *_acts++ )
		{
	case 0:
#line 366 "lib/eolian/eo_lexer.rl"
	{
      toknz->current_line += 1;
      DBG("inc[%d] %d", toknz->cs, toknz->current_line);
   }
	break;
	case 1:
#line 371 "lib/eolian/eo_lexer.rl"
	{
      toknz->saved.line = toknz->current_line;
      DBG("save line[%d] %d", toknz->cs, toknz->current_line);
   }
	break;
	case 2:
#line 376 "lib/eolian/eo_lexer.rl"
	{
      toknz->saved.tok = ( toknz->p);
      DBG("save token[%d] %p %c", toknz->cs, ( toknz->p), *( toknz->p));
   }
	break;
	case 3:
#line 381 "lib/eolian/eo_lexer.rl"
	{
      DBG("move ts %d chars forward", (int)(( toknz->p) - toknz->ts));
      toknz->ts = ( toknz->p);
   }
	break;
	case 4:
#line 460 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.accessor) ABORT(toknz, "No accessor!!!");
      if (toknz->tmp.accessor->ret != NULL)
        ABORT(toknz, "accessor has already a return type");
      toknz->tmp.accessor->ret = _eo_tokenizer_return_get(toknz, ( toknz->p));
   }
	break;
	case 5:
#line 467 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.accessor) ABORT(toknz, "No accessor!!!");
      if (!toknz->tmp.accessor->ret) ABORT(toknz, "No ret!!!");
      if (toknz->tmp.accessor->ret->comment != NULL)
        ABORT(toknz, "accessor return type has already a comment");
      toknz->tmp.accessor->ret->comment = _eo_tokenizer_token_get(toknz, ( toknz->p)-2);
      INF("        %s", toknz->tmp.accessor->ret->comment);
   }
	break;
	case 6:
#line 476 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.accessor) ABORT(toknz, "No accessor!!!");
      toknz->tmp.accessor->legacy = _eo_tokenizer_token_get(toknz, ( toknz->p));
   }
	break;
	case 7:
#line 490 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.accessor_param = _eo_tokenizer_accessor_param_get(toknz, ( toknz->p));
   }
	break;
	case 8:
#line 494 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.accessor_param)
         ABORT(toknz, "No accessor param!!!");
      toknz->tmp.accessor_param->attrs = _eo_tokenizer_token_get(toknz, ( toknz->p));
      toknz->tmp.accessor->params =
         eina_list_append(toknz->tmp.accessor->params, toknz->tmp.accessor_param);
      toknz->tmp.accessor_param = NULL;
   }
	break;
	case 9:
#line 523 "lib/eolian/eo_lexer.rl"
	{
      const char *c = _eo_tokenizer_token_get(toknz, ( toknz->p)-2);
      if (toknz->tmp.param == NULL)
        ABORT(toknz, "no parameter set to associate this comment to: %s", c);
      toknz->tmp.param->comment = c;
      toknz->tmp.param = NULL;
   }
	break;
	case 10:
#line 531 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.param = _eo_tokenizer_param_get(toknz, ( toknz->p));
      if (toknz->tmp.params)
        *(toknz->tmp.params) = eina_list_append(*(toknz->tmp.params), toknz->tmp.param);
      else
        ABORT(toknz, "got a param but there is no property nor method waiting for it");
      INF("        %s : %s", toknz->tmp.param->name, toknz->tmp.param->type);
   }
	break;
	case 11:
#line 630 "lib/eolian/eo_lexer.rl"
	{
      if (toknz->tmp.prop != NULL)
        ABORT(toknz, "there is a pending property definition %s", toknz->tmp.prop->name);
      toknz->tmp.prop = _eo_tokenizer_property_get(toknz, ( toknz->p));
   }
	break;
	case 12:
#line 636 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.fscope = _eo_tokenizer_scope_get(toknz, ( toknz->p));
   }
	break;
	case 13:
#line 674 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      if (toknz->tmp.meth->ret != NULL)
        ABORT(toknz, "method '%s' has already a return type", toknz->tmp.meth->name);
      toknz->tmp.meth->ret = _eo_tokenizer_return_get(toknz, ( toknz->p));
   }
	break;
	case 14:
#line 681 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      if (toknz->tmp.meth->ret == NULL) ABORT(toknz, "No ret!!!");
      if (toknz->tmp.meth->ret->comment != NULL)
        ABORT(toknz, "method '%s' return type has already a comment", toknz->tmp.meth->name);
      toknz->tmp.meth->ret->comment = _eo_tokenizer_token_get(toknz, ( toknz->p)-2);
      INF("        %s", toknz->tmp.meth->ret->comment);
   }
	break;
	case 15:
#line 690 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      toknz->tmp.meth->legacy = _eo_tokenizer_token_get(toknz, ( toknz->p));
   }
	break;
	case 16:
#line 695 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      toknz->tmp.meth->obj_const = EINA_TRUE;
      INF("        obj const");
   }
	break;
	case 17:
#line 753 "lib/eolian/eo_lexer.rl"
	{
      if (toknz->tmp.meth != NULL)
        ABORT(toknz, "there is a pending method definition %s", toknz->tmp.meth->name);
      toknz->tmp.meth = _eo_tokenizer_method_get(toknz, ( toknz->p));
   }
	break;
	case 18:
#line 759 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.fscope = _eo_tokenizer_scope_get(toknz, ( toknz->p));
   }
	break;
	case 19:
#line 789 "lib/eolian/eo_lexer.rl"
	{
      const char *base = _eo_tokenizer_token_get(toknz, ( toknz->p));
      toknz->tmp.str_items = eina_list_append(toknz->tmp.str_items, base);
   }
	break;
	case 20:
#line 794 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      toknz->tmp.kls->inherits = toknz->tmp.str_items;
      toknz->tmp.str_items = NULL;
   }
	break;
	case 21:
#line 838 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      toknz->tmp.event = _eo_tokenizer_event_get(toknz, ( toknz->p));
      toknz->tmp.kls->events = eina_list_append(toknz->tmp.kls->events, toknz->tmp.event);
   }
	break;
	case 22:
#line 844 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.event) ABORT(toknz, "No event!!!");
      if (toknz->tmp.event->type != NULL)
        ABORT(toknz, "event %s has already a type %s", toknz->tmp.event->name, toknz->tmp.event->type);
      toknz->tmp.event->type = _eo_tokenizer_token_get(toknz, ( toknz->p)-1);
   }
	break;
	case 23:
#line 851 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.event) ABORT(toknz, "No event!!!");
      if (toknz->tmp.event->comment != NULL)
        ABORT(toknz, "event %s has already a comment", toknz->tmp.event->name);
      toknz->tmp.event->comment = _eo_tokenizer_token_get(toknz, ( toknz->p)-2);
      toknz->tmp.event = NULL;
   }
	break;
	case 24:
#line 859 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      if (toknz->tmp.kls->legacy_prefix != NULL)
        ABORT(toknz, "A legacy prefix has already been given");
      toknz->tmp.kls->legacy_prefix = _eo_tokenizer_token_get(toknz, ( toknz->p));
   }
	break;
	case 25:
#line 868 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      if (toknz->tmp.kls->eo_prefix != NULL)
        ABORT(toknz, "An Eo prefix has already been given");
      toknz->tmp.kls->eo_prefix = _eo_tokenizer_token_get(toknz, ( toknz->p));
   }
	break;
	case 26:
#line 877 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      if (toknz->tmp.kls->data_type != NULL)
        ABORT(toknz, "A data type has already been given");
      toknz->tmp.kls->data_type = _eo_tokenizer_token_get(toknz, ( toknz->p));
   }
	break;
	case 27:
#line 890 "lib/eolian/eo_lexer.rl"
	{
        if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
        toknz->tmp.impl = _eo_tokenizer_implement_get(toknz, ( toknz->p));
        toknz->tmp.kls->implements = eina_list_append(toknz->tmp.kls->implements, toknz->tmp.impl);
   }
	break;
	case 28:
#line 934 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.kls_type = EOLIAN_CLASS_REGULAR;
   }
	break;
	case 29:
#line 937 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.kls_type = EOLIAN_CLASS_ABSTRACT;
   }
	break;
	case 30:
#line 940 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.kls_type = EOLIAN_CLASS_MIXIN;
   }
	break;
	case 31:
#line 943 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.kls_type = EOLIAN_CLASS_INTERFACE;
   }
	break;
	case 32:
#line 947 "lib/eolian/eo_lexer.rl"
	{
      if (toknz->tmp.kls != NULL)
        ABORT(toknz, "there is a pending class definition %s", toknz->tmp.kls->name);
      toknz->tmp.kls = _eo_tokenizer_class_get(toknz, ( toknz->p));
      toknz->tmp.kls->type = toknz->tmp.kls_type;
   }
	break;
	case 35:
#line 1 "NONE"
	{ toknz->te = ( toknz->p)+1;}
	break;
	case 36:
#line 452 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.accessor) ABORT(toknz, "No accessor!!!");
      if (toknz->tmp.accessor->comment != NULL)
        ABORT(toknz, "accessor has already a comment");
      toknz->tmp.accessor->comment = _eo_tokenizer_token_get(toknz, ( toknz->p)-1);
      INF("        %s", toknz->tmp.accessor->comment);
   }}
	break;
	case 37:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 38:
#line 515 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;}
	break;
	case 39:
#line 516 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;}
	break;
	case 40:
#line 481 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("      }");
      if (!toknz->tmp.prop) ABORT(toknz, "No prop!!!");
      toknz->tmp.prop->accessors = eina_list_append(toknz->tmp.prop->accessors, toknz->tmp.accessor);
      toknz->tmp.accessor = NULL;
      toknz->current_nesting--;
      { toknz->cs = 315; goto _again;}
   }}
	break;
	case 41:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 42:
#line 511 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 43:
#line 514 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 44:
#line 481 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      INF("      }");
      if (!toknz->tmp.prop) ABORT(toknz, "No prop!!!");
      toknz->tmp.prop->accessors = eina_list_append(toknz->tmp.prop->accessors, toknz->tmp.accessor);
      toknz->tmp.accessor = NULL;
      toknz->current_nesting--;
      { toknz->cs = 315; goto _again;}
   }}
	break;
	case 45:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 46:
#line 514 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}}
	break;
	case 47:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 48:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 49:
#line 540 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("      }");
      toknz->tmp.param = NULL;
      toknz->current_nesting--;
      if (toknz->tmp.prop)
        { toknz->cs = 315; goto _again;}
      else if (toknz->tmp.meth)
        { toknz->cs = 329; goto _again;}
      else
        ABORT(toknz, "leaving tokenize_params but there is no property nor method pending");
   }}
	break;
	case 50:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 51:
#line 556 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 52:
#line 558 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 53:
#line 540 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      INF("      }");
      toknz->tmp.param = NULL;
      toknz->current_nesting--;
      if (toknz->tmp.prop)
        { toknz->cs = 315; goto _again;}
      else if (toknz->tmp.meth)
        { toknz->cs = 329; goto _again;}
      else
        ABORT(toknz, "leaving tokenize_params but there is no property nor method pending");
   }}
	break;
	case 54:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 55:
#line 558 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}}
	break;
	case 56:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 57:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 58:
#line 565 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("      get {");
      toknz->tmp.accessor = _eo_tokenizer_accessor_get(toknz, GETTER);
      toknz->current_nesting++;
      { toknz->cs = 299; goto _again;}
   }}
	break;
	case 59:
#line 572 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("      set {");
      toknz->tmp.accessor = _eo_tokenizer_accessor_get(toknz, SETTER);
      toknz->current_nesting++;
      { toknz->cs = 299; goto _again;}
   }}
	break;
	case 60:
#line 579 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("      keys {");
      toknz->current_nesting++;
      toknz->tmp.params = &(toknz->tmp.prop->keys);
      { toknz->cs = 308; goto _again;}
   }}
	break;
	case 61:
#line 586 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("      values {");
      toknz->current_nesting++;
      toknz->tmp.params = &(toknz->tmp.prop->values);
      { toknz->cs = 308; goto _again;}
   }}
	break;
	case 62:
#line 593 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.prop) ABORT(toknz, "No property!!!");
      if (eina_list_count(toknz->tmp.prop->accessors) == 0)
        WRN("property '%s' has no accessors.", toknz->tmp.prop->name);
      INF("    }");
      toknz->tmp.kls->properties = eina_list_append(toknz->tmp.kls->properties, toknz->tmp.prop);
      toknz->tmp.prop = NULL;
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting--;
      { toknz->cs = 323; goto _again;}
   }}
	break;
	case 63:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 64:
#line 611 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 65:
#line 593 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      if (!toknz->tmp.prop) ABORT(toknz, "No property!!!");
      if (eina_list_count(toknz->tmp.prop->accessors) == 0)
        WRN("property '%s' has no accessors.", toknz->tmp.prop->name);
      INF("    }");
      toknz->tmp.kls->properties = eina_list_append(toknz->tmp.kls->properties, toknz->tmp.prop);
      toknz->tmp.prop = NULL;
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting--;
      { toknz->cs = 323; goto _again;}
   }}
	break;
	case 66:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 67:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 68:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 69:
#line 623 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.prop) ABORT(toknz, "No property!!!");
      INF("    %s {", toknz->tmp.prop->name);
      toknz->current_nesting++;
      { toknz->cs = 315; goto _again;}
   }}
	break;
	case 70:
#line 640 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("  }");
      toknz->current_nesting--;
      { toknz->cs = 345; goto _again;}
   }}
	break;
	case 71:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 72:
#line 649 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 73:
#line 640 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      INF("  }");
      toknz->current_nesting--;
      { toknz->cs = 345; goto _again;}
   }}
	break;
	case 74:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 75:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 76:
#line 658 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      if (toknz->tmp.meth->comment != NULL)
        ABORT(toknz, "method has already a comment");
      toknz->tmp.meth->comment = _eo_tokenizer_token_get(toknz, ( toknz->p)-1);
      INF("        %s", toknz->tmp.meth->comment);
   }}
	break;
	case 77:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 78:
#line 666 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      INF("      params {");
      toknz->current_nesting++;
      toknz->tmp.params = &(toknz->tmp.meth->params);
      { toknz->cs = 308; goto _again;}
   }}
	break;
	case 79:
#line 738 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;}
	break;
	case 80:
#line 739 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;}
	break;
	case 81:
#line 701 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      Eina_List **l = NULL;
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      INF("    }");
      switch (toknz->current_methods_type) {
        case METH_CONSTRUCTOR:
          l = &toknz->tmp.kls->constructors;
          break;
        case METH_REGULAR:
          l = &toknz->tmp.kls->methods;
          break;
        default:
          ABORT(toknz, "unknown method type %d", toknz->current_methods_type);
      }
      toknz->tmp.meth->type = toknz->current_methods_type;
      *l = eina_list_append(*l, toknz->tmp.meth);
      toknz->tmp.meth = NULL;
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting--;
      { toknz->cs = 339; goto _again;}
   }}
	break;
	case 82:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 83:
#line 733 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 84:
#line 737 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 85:
#line 701 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      Eina_List **l = NULL;
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      INF("    }");
      switch (toknz->current_methods_type) {
        case METH_CONSTRUCTOR:
          l = &toknz->tmp.kls->constructors;
          break;
        case METH_REGULAR:
          l = &toknz->tmp.kls->methods;
          break;
        default:
          ABORT(toknz, "unknown method type %d", toknz->current_methods_type);
      }
      toknz->tmp.meth->type = toknz->current_methods_type;
      *l = eina_list_append(*l, toknz->tmp.meth);
      toknz->tmp.meth = NULL;
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting--;
      { toknz->cs = 339; goto _again;}
   }}
	break;
	case 86:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 87:
#line 737 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}}
	break;
	case 88:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 89:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 90:
#line 746 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      INF("    %s {", toknz->tmp.meth->name);
      toknz->current_nesting++;
      { toknz->cs = 329; goto _again;}
   }}
	break;
	case 91:
#line 763 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("  }");
      toknz->current_methods_type = METH_TYPE_LAST;
      toknz->current_nesting--;
      { toknz->cs = 345; goto _again;}
   }}
	break;
	case 92:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 93:
#line 773 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 94:
#line 763 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      INF("  }");
      toknz->current_methods_type = METH_TYPE_LAST;
      toknz->current_nesting--;
      { toknz->cs = 345; goto _again;}
   }}
	break;
	case 95:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 96:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 97:
#line 782 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      if (toknz->tmp.kls->comment != NULL)
        ABORT(toknz, "class %s has already a comment", toknz->tmp.kls->name);
      toknz->tmp.kls->comment = _eo_tokenizer_token_get(toknz, ( toknz->p)-1);
   }}
	break;
	case 98:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 99:
#line 800 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
   }}
	break;
	case 100:
#line 803 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
   }}
	break;
	case 101:
#line 806 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("  constructors {");
      toknz->current_methods_type = METH_CONSTRUCTOR;
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting++;
      { toknz->cs = 339; goto _again;}
   }}
	break;
	case 102:
#line 814 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("  properties {");
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting++;
      { toknz->cs = 323; goto _again;}
   }}
	break;
	case 103:
#line 821 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("  begin methods");
      toknz->current_methods_type = METH_REGULAR;
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting++;
      { toknz->cs = 339; goto _again;}
   }}
	break;
	case 104:
#line 829 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      INF("end class: %s", toknz->tmp.kls->name);
      toknz->classes = eina_list_append(toknz->classes, toknz->tmp.kls);
      toknz->tmp.kls = NULL;
      toknz->current_nesting--;
      { toknz->cs = 292; goto _again;}
   }}
	break;
	case 105:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 106:
#line 910 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 107:
#line 913 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 108:
#line 914 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 109:
#line 915 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 110:
#line 800 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
   }}
	break;
	case 111:
#line 803 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
   }}
	break;
	case 112:
#line 829 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      INF("end class: %s", toknz->tmp.kls->name);
      toknz->classes = eina_list_append(toknz->classes, toknz->tmp.kls);
      toknz->tmp.kls = NULL;
      toknz->current_nesting--;
      { toknz->cs = 292; goto _again;}
   }}
	break;
	case 113:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 114:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 115:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 116:
#line 927 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      INF("begin class: %s", toknz->tmp.kls->name);
      toknz->current_nesting++;
      { toknz->cs = 345; goto _again;}
   }}
	break;
	case 117:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 118:
#line 962 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 119:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 120:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
#line 2719 "lib/eolian/eo_lexer.c"
		}
	}

_again:
	_acts = _eo_tokenizer_actions + _eo_tokenizer_to_state_actions[ toknz->cs];
	_nacts = (unsigned int) *_acts++;
	while ( _nacts-- > 0 ) {
		switch ( *_acts++ ) {
	case 33:
#line 1 "NONE"
	{ toknz->ts = 0;}
	break;
#line 2732 "lib/eolian/eo_lexer.c"
		}
	}

	if ( ++( toknz->p) != ( toknz->pe) )
		goto _resume;
	_test_eof: {}
	if ( ( toknz->p) == ( toknz->eof) )
	{
	if ( _eo_tokenizer_eof_trans[ toknz->cs] > 0 ) {
		_trans = _eo_tokenizer_eof_trans[ toknz->cs] - 1;
		goto _eof_trans;
	}
	}

	_out: {}
	}

#line 1017 "lib/eolian/eo_lexer.rl"

        if ( toknz->cs == 
#line 2753 "lib/eolian/eo_lexer.c"
-1
#line 1018 "lib/eolian/eo_lexer.rl"
 )
          {
             ERR("%s: wrong termination", source);
             ret = EINA_FALSE;
             break;
          }

        if ( toknz->ts == 0 )
          have = 0;
        else
          {
             DBG("move data and pointers before buffer feed");
             have = toknz->pe - toknz->ts;
             offset = toknz->ts - toknz->buf;
             memmove(toknz->buf, toknz->ts, have);
             toknz->te -= offset;
             toknz->ts = toknz->buf;
          }

        if (toknz->saved.tok != NULL)
          {
             if ((have == 0) || ((toknz->saved.tok - offset) < toknz->buf))
               {
                  WRN("reset lost saved token %p", toknz->saved.tok);
                  toknz->saved.tok = NULL;
               }
             else
               toknz->saved.tok -= offset;
          }
     }

   fclose(stream);

   return ret;
}

static Eina_Bool
eo_tokenizer_mem_walk(Eo_Tokenizer *toknz, const char *source, char *buffer, unsigned int len)
{
   INF("tokenize %s...", source);
   toknz->source = eina_stringshare_add(source);

   Eina_Bool ret = EINA_TRUE;

   
#line 2801 "lib/eolian/eo_lexer.c"
	{
	 toknz->cs = eo_tokenizer_start;
	 toknz->ts = 0;
	 toknz->te = 0;
	 toknz->act = 0;
	}

#line 1063 "lib/eolian/eo_lexer.rl"

   toknz->p = buffer;

   toknz->pe = toknz->p + len;

   toknz->eof = toknz->pe;

   
#line 2818 "lib/eolian/eo_lexer.c"
	{
	int _klen;
	unsigned int _trans;
	const char *_acts;
	unsigned int _nacts;
	const char *_keys;

	if ( ( toknz->p) == ( toknz->pe) )
		goto _test_eof;
_resume:
	_acts = _eo_tokenizer_actions + _eo_tokenizer_from_state_actions[ toknz->cs];
	_nacts = (unsigned int) *_acts++;
	while ( _nacts-- > 0 ) {
		switch ( *_acts++ ) {
	case 34:
#line 1 "NONE"
	{ toknz->ts = ( toknz->p);}
	break;
#line 2837 "lib/eolian/eo_lexer.c"
		}
	}

	_keys = _eo_tokenizer_trans_keys + _eo_tokenizer_key_offsets[ toknz->cs];
	_trans = _eo_tokenizer_index_offsets[ toknz->cs];

	_klen = _eo_tokenizer_single_lengths[ toknz->cs];
	if ( _klen > 0 ) {
		const char *_lower = _keys;
		const char *_mid;
		const char *_upper = _keys + _klen - 1;
		while (1) {
			if ( _upper < _lower )
				break;

			_mid = _lower + ((_upper-_lower) >> 1);
			if ( (*( toknz->p)) < *_mid )
				_upper = _mid - 1;
			else if ( (*( toknz->p)) > *_mid )
				_lower = _mid + 1;
			else {
				_trans += (unsigned int)(_mid - _keys);
				goto _match;
			}
		}
		_keys += _klen;
		_trans += _klen;
	}

	_klen = _eo_tokenizer_range_lengths[ toknz->cs];
	if ( _klen > 0 ) {
		const char *_lower = _keys;
		const char *_mid;
		const char *_upper = _keys + (_klen<<1) - 2;
		while (1) {
			if ( _upper < _lower )
				break;

			_mid = _lower + (((_upper-_lower) >> 1) & ~1);
			if ( (*( toknz->p)) < _mid[0] )
				_upper = _mid - 2;
			else if ( (*( toknz->p)) > _mid[1] )
				_lower = _mid + 2;
			else {
				_trans += (unsigned int)((_mid - _keys)>>1);
				goto _match;
			}
		}
		_trans += _klen;
	}

_match:
	_trans = _eo_tokenizer_indicies[_trans];
_eof_trans:
	 toknz->cs = _eo_tokenizer_trans_targs[_trans];

	if ( _eo_tokenizer_trans_actions[_trans] == 0 )
		goto _again;

	_acts = _eo_tokenizer_actions + _eo_tokenizer_trans_actions[_trans];
	_nacts = (unsigned int) *_acts++;
	while ( _nacts-- > 0 )
	{
		switch ( *_acts++ )
		{
	case 0:
#line 366 "lib/eolian/eo_lexer.rl"
	{
      toknz->current_line += 1;
      DBG("inc[%d] %d", toknz->cs, toknz->current_line);
   }
	break;
	case 1:
#line 371 "lib/eolian/eo_lexer.rl"
	{
      toknz->saved.line = toknz->current_line;
      DBG("save line[%d] %d", toknz->cs, toknz->current_line);
   }
	break;
	case 2:
#line 376 "lib/eolian/eo_lexer.rl"
	{
      toknz->saved.tok = ( toknz->p);
      DBG("save token[%d] %p %c", toknz->cs, ( toknz->p), *( toknz->p));
   }
	break;
	case 3:
#line 381 "lib/eolian/eo_lexer.rl"
	{
      DBG("move ts %d chars forward", (int)(( toknz->p) - toknz->ts));
      toknz->ts = ( toknz->p);
   }
	break;
	case 4:
#line 460 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.accessor) ABORT(toknz, "No accessor!!!");
      if (toknz->tmp.accessor->ret != NULL)
        ABORT(toknz, "accessor has already a return type");
      toknz->tmp.accessor->ret = _eo_tokenizer_return_get(toknz, ( toknz->p));
   }
	break;
	case 5:
#line 467 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.accessor) ABORT(toknz, "No accessor!!!");
      if (!toknz->tmp.accessor->ret) ABORT(toknz, "No ret!!!");
      if (toknz->tmp.accessor->ret->comment != NULL)
        ABORT(toknz, "accessor return type has already a comment");
      toknz->tmp.accessor->ret->comment = _eo_tokenizer_token_get(toknz, ( toknz->p)-2);
      INF("        %s", toknz->tmp.accessor->ret->comment);
   }
	break;
	case 6:
#line 476 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.accessor) ABORT(toknz, "No accessor!!!");
      toknz->tmp.accessor->legacy = _eo_tokenizer_token_get(toknz, ( toknz->p));
   }
	break;
	case 7:
#line 490 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.accessor_param = _eo_tokenizer_accessor_param_get(toknz, ( toknz->p));
   }
	break;
	case 8:
#line 494 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.accessor_param)
         ABORT(toknz, "No accessor param!!!");
      toknz->tmp.accessor_param->attrs = _eo_tokenizer_token_get(toknz, ( toknz->p));
      toknz->tmp.accessor->params =
         eina_list_append(toknz->tmp.accessor->params, toknz->tmp.accessor_param);
      toknz->tmp.accessor_param = NULL;
   }
	break;
	case 9:
#line 523 "lib/eolian/eo_lexer.rl"
	{
      const char *c = _eo_tokenizer_token_get(toknz, ( toknz->p)-2);
      if (toknz->tmp.param == NULL)
        ABORT(toknz, "no parameter set to associate this comment to: %s", c);
      toknz->tmp.param->comment = c;
      toknz->tmp.param = NULL;
   }
	break;
	case 10:
#line 531 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.param = _eo_tokenizer_param_get(toknz, ( toknz->p));
      if (toknz->tmp.params)
        *(toknz->tmp.params) = eina_list_append(*(toknz->tmp.params), toknz->tmp.param);
      else
        ABORT(toknz, "got a param but there is no property nor method waiting for it");
      INF("        %s : %s", toknz->tmp.param->name, toknz->tmp.param->type);
   }
	break;
	case 11:
#line 630 "lib/eolian/eo_lexer.rl"
	{
      if (toknz->tmp.prop != NULL)
        ABORT(toknz, "there is a pending property definition %s", toknz->tmp.prop->name);
      toknz->tmp.prop = _eo_tokenizer_property_get(toknz, ( toknz->p));
   }
	break;
	case 12:
#line 636 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.fscope = _eo_tokenizer_scope_get(toknz, ( toknz->p));
   }
	break;
	case 13:
#line 674 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      if (toknz->tmp.meth->ret != NULL)
        ABORT(toknz, "method '%s' has already a return type", toknz->tmp.meth->name);
      toknz->tmp.meth->ret = _eo_tokenizer_return_get(toknz, ( toknz->p));
   }
	break;
	case 14:
#line 681 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      if (toknz->tmp.meth->ret == NULL) ABORT(toknz, "No ret!!!");
      if (toknz->tmp.meth->ret->comment != NULL)
        ABORT(toknz, "method '%s' return type has already a comment", toknz->tmp.meth->name);
      toknz->tmp.meth->ret->comment = _eo_tokenizer_token_get(toknz, ( toknz->p)-2);
      INF("        %s", toknz->tmp.meth->ret->comment);
   }
	break;
	case 15:
#line 690 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      toknz->tmp.meth->legacy = _eo_tokenizer_token_get(toknz, ( toknz->p));
   }
	break;
	case 16:
#line 695 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      toknz->tmp.meth->obj_const = EINA_TRUE;
      INF("        obj const");
   }
	break;
	case 17:
#line 753 "lib/eolian/eo_lexer.rl"
	{
      if (toknz->tmp.meth != NULL)
        ABORT(toknz, "there is a pending method definition %s", toknz->tmp.meth->name);
      toknz->tmp.meth = _eo_tokenizer_method_get(toknz, ( toknz->p));
   }
	break;
	case 18:
#line 759 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.fscope = _eo_tokenizer_scope_get(toknz, ( toknz->p));
   }
	break;
	case 19:
#line 789 "lib/eolian/eo_lexer.rl"
	{
      const char *base = _eo_tokenizer_token_get(toknz, ( toknz->p));
      toknz->tmp.str_items = eina_list_append(toknz->tmp.str_items, base);
   }
	break;
	case 20:
#line 794 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      toknz->tmp.kls->inherits = toknz->tmp.str_items;
      toknz->tmp.str_items = NULL;
   }
	break;
	case 21:
#line 838 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      toknz->tmp.event = _eo_tokenizer_event_get(toknz, ( toknz->p));
      toknz->tmp.kls->events = eina_list_append(toknz->tmp.kls->events, toknz->tmp.event);
   }
	break;
	case 22:
#line 844 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.event) ABORT(toknz, "No event!!!");
      if (toknz->tmp.event->type != NULL)
        ABORT(toknz, "event %s has already a type %s", toknz->tmp.event->name, toknz->tmp.event->type);
      toknz->tmp.event->type = _eo_tokenizer_token_get(toknz, ( toknz->p)-1);
   }
	break;
	case 23:
#line 851 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.event) ABORT(toknz, "No event!!!");
      if (toknz->tmp.event->comment != NULL)
        ABORT(toknz, "event %s has already a comment", toknz->tmp.event->name);
      toknz->tmp.event->comment = _eo_tokenizer_token_get(toknz, ( toknz->p)-2);
      toknz->tmp.event = NULL;
   }
	break;
	case 24:
#line 859 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      if (toknz->tmp.kls->legacy_prefix != NULL)
        ABORT(toknz, "A legacy prefix has already been given");
      toknz->tmp.kls->legacy_prefix = _eo_tokenizer_token_get(toknz, ( toknz->p));
   }
	break;
	case 25:
#line 868 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      if (toknz->tmp.kls->eo_prefix != NULL)
        ABORT(toknz, "An Eo prefix has already been given");
      toknz->tmp.kls->eo_prefix = _eo_tokenizer_token_get(toknz, ( toknz->p));
   }
	break;
	case 26:
#line 877 "lib/eolian/eo_lexer.rl"
	{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      if (toknz->tmp.kls->data_type != NULL)
        ABORT(toknz, "A data type has already been given");
      toknz->tmp.kls->data_type = _eo_tokenizer_token_get(toknz, ( toknz->p));
   }
	break;
	case 27:
#line 890 "lib/eolian/eo_lexer.rl"
	{
        if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
        toknz->tmp.impl = _eo_tokenizer_implement_get(toknz, ( toknz->p));
        toknz->tmp.kls->implements = eina_list_append(toknz->tmp.kls->implements, toknz->tmp.impl);
   }
	break;
	case 28:
#line 934 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.kls_type = EOLIAN_CLASS_REGULAR;
   }
	break;
	case 29:
#line 937 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.kls_type = EOLIAN_CLASS_ABSTRACT;
   }
	break;
	case 30:
#line 940 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.kls_type = EOLIAN_CLASS_MIXIN;
   }
	break;
	case 31:
#line 943 "lib/eolian/eo_lexer.rl"
	{
      toknz->tmp.kls_type = EOLIAN_CLASS_INTERFACE;
   }
	break;
	case 32:
#line 947 "lib/eolian/eo_lexer.rl"
	{
      if (toknz->tmp.kls != NULL)
        ABORT(toknz, "there is a pending class definition %s", toknz->tmp.kls->name);
      toknz->tmp.kls = _eo_tokenizer_class_get(toknz, ( toknz->p));
      toknz->tmp.kls->type = toknz->tmp.kls_type;
   }
	break;
	case 35:
#line 1 "NONE"
	{ toknz->te = ( toknz->p)+1;}
	break;
	case 36:
#line 452 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.accessor) ABORT(toknz, "No accessor!!!");
      if (toknz->tmp.accessor->comment != NULL)
        ABORT(toknz, "accessor has already a comment");
      toknz->tmp.accessor->comment = _eo_tokenizer_token_get(toknz, ( toknz->p)-1);
      INF("        %s", toknz->tmp.accessor->comment);
   }}
	break;
	case 37:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 38:
#line 515 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;}
	break;
	case 39:
#line 516 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;}
	break;
	case 40:
#line 481 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("      }");
      if (!toknz->tmp.prop) ABORT(toknz, "No prop!!!");
      toknz->tmp.prop->accessors = eina_list_append(toknz->tmp.prop->accessors, toknz->tmp.accessor);
      toknz->tmp.accessor = NULL;
      toknz->current_nesting--;
      { toknz->cs = 315; goto _again;}
   }}
	break;
	case 41:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 42:
#line 511 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 43:
#line 514 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 44:
#line 481 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      INF("      }");
      if (!toknz->tmp.prop) ABORT(toknz, "No prop!!!");
      toknz->tmp.prop->accessors = eina_list_append(toknz->tmp.prop->accessors, toknz->tmp.accessor);
      toknz->tmp.accessor = NULL;
      toknz->current_nesting--;
      { toknz->cs = 315; goto _again;}
   }}
	break;
	case 45:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 46:
#line 514 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}}
	break;
	case 47:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 48:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 49:
#line 540 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("      }");
      toknz->tmp.param = NULL;
      toknz->current_nesting--;
      if (toknz->tmp.prop)
        { toknz->cs = 315; goto _again;}
      else if (toknz->tmp.meth)
        { toknz->cs = 329; goto _again;}
      else
        ABORT(toknz, "leaving tokenize_params but there is no property nor method pending");
   }}
	break;
	case 50:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 51:
#line 556 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 52:
#line 558 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 53:
#line 540 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      INF("      }");
      toknz->tmp.param = NULL;
      toknz->current_nesting--;
      if (toknz->tmp.prop)
        { toknz->cs = 315; goto _again;}
      else if (toknz->tmp.meth)
        { toknz->cs = 329; goto _again;}
      else
        ABORT(toknz, "leaving tokenize_params but there is no property nor method pending");
   }}
	break;
	case 54:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 55:
#line 558 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}}
	break;
	case 56:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 57:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 58:
#line 565 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("      get {");
      toknz->tmp.accessor = _eo_tokenizer_accessor_get(toknz, GETTER);
      toknz->current_nesting++;
      { toknz->cs = 299; goto _again;}
   }}
	break;
	case 59:
#line 572 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("      set {");
      toknz->tmp.accessor = _eo_tokenizer_accessor_get(toknz, SETTER);
      toknz->current_nesting++;
      { toknz->cs = 299; goto _again;}
   }}
	break;
	case 60:
#line 579 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("      keys {");
      toknz->current_nesting++;
      toknz->tmp.params = &(toknz->tmp.prop->keys);
      { toknz->cs = 308; goto _again;}
   }}
	break;
	case 61:
#line 586 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("      values {");
      toknz->current_nesting++;
      toknz->tmp.params = &(toknz->tmp.prop->values);
      { toknz->cs = 308; goto _again;}
   }}
	break;
	case 62:
#line 593 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.prop) ABORT(toknz, "No property!!!");
      if (eina_list_count(toknz->tmp.prop->accessors) == 0)
        WRN("property '%s' has no accessors.", toknz->tmp.prop->name);
      INF("    }");
      toknz->tmp.kls->properties = eina_list_append(toknz->tmp.kls->properties, toknz->tmp.prop);
      toknz->tmp.prop = NULL;
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting--;
      { toknz->cs = 323; goto _again;}
   }}
	break;
	case 63:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 64:
#line 611 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 65:
#line 593 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      if (!toknz->tmp.prop) ABORT(toknz, "No property!!!");
      if (eina_list_count(toknz->tmp.prop->accessors) == 0)
        WRN("property '%s' has no accessors.", toknz->tmp.prop->name);
      INF("    }");
      toknz->tmp.kls->properties = eina_list_append(toknz->tmp.kls->properties, toknz->tmp.prop);
      toknz->tmp.prop = NULL;
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting--;
      { toknz->cs = 323; goto _again;}
   }}
	break;
	case 66:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 67:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 68:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 69:
#line 623 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.prop) ABORT(toknz, "No property!!!");
      INF("    %s {", toknz->tmp.prop->name);
      toknz->current_nesting++;
      { toknz->cs = 315; goto _again;}
   }}
	break;
	case 70:
#line 640 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("  }");
      toknz->current_nesting--;
      { toknz->cs = 345; goto _again;}
   }}
	break;
	case 71:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 72:
#line 649 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 73:
#line 640 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      INF("  }");
      toknz->current_nesting--;
      { toknz->cs = 345; goto _again;}
   }}
	break;
	case 74:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 75:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 76:
#line 658 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      if (toknz->tmp.meth->comment != NULL)
        ABORT(toknz, "method has already a comment");
      toknz->tmp.meth->comment = _eo_tokenizer_token_get(toknz, ( toknz->p)-1);
      INF("        %s", toknz->tmp.meth->comment);
   }}
	break;
	case 77:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 78:
#line 666 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      INF("      params {");
      toknz->current_nesting++;
      toknz->tmp.params = &(toknz->tmp.meth->params);
      { toknz->cs = 308; goto _again;}
   }}
	break;
	case 79:
#line 738 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;}
	break;
	case 80:
#line 739 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;}
	break;
	case 81:
#line 701 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      Eina_List **l = NULL;
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      INF("    }");
      switch (toknz->current_methods_type) {
        case METH_CONSTRUCTOR:
          l = &toknz->tmp.kls->constructors;
          break;
        case METH_REGULAR:
          l = &toknz->tmp.kls->methods;
          break;
        default:
          ABORT(toknz, "unknown method type %d", toknz->current_methods_type);
      }
      toknz->tmp.meth->type = toknz->current_methods_type;
      *l = eina_list_append(*l, toknz->tmp.meth);
      toknz->tmp.meth = NULL;
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting--;
      { toknz->cs = 339; goto _again;}
   }}
	break;
	case 82:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 83:
#line 733 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 84:
#line 737 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 85:
#line 701 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      Eina_List **l = NULL;
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      INF("    }");
      switch (toknz->current_methods_type) {
        case METH_CONSTRUCTOR:
          l = &toknz->tmp.kls->constructors;
          break;
        case METH_REGULAR:
          l = &toknz->tmp.kls->methods;
          break;
        default:
          ABORT(toknz, "unknown method type %d", toknz->current_methods_type);
      }
      toknz->tmp.meth->type = toknz->current_methods_type;
      *l = eina_list_append(*l, toknz->tmp.meth);
      toknz->tmp.meth = NULL;
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting--;
      { toknz->cs = 339; goto _again;}
   }}
	break;
	case 86:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 87:
#line 737 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}}
	break;
	case 88:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 89:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 90:
#line 746 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.meth) ABORT(toknz, "No method!!!");
      INF("    %s {", toknz->tmp.meth->name);
      toknz->current_nesting++;
      { toknz->cs = 329; goto _again;}
   }}
	break;
	case 91:
#line 763 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("  }");
      toknz->current_methods_type = METH_TYPE_LAST;
      toknz->current_nesting--;
      { toknz->cs = 345; goto _again;}
   }}
	break;
	case 92:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 93:
#line 773 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 94:
#line 763 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      INF("  }");
      toknz->current_methods_type = METH_TYPE_LAST;
      toknz->current_nesting--;
      { toknz->cs = 345; goto _again;}
   }}
	break;
	case 95:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 96:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 97:
#line 782 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      if (toknz->tmp.kls->comment != NULL)
        ABORT(toknz, "class %s has already a comment", toknz->tmp.kls->name);
      toknz->tmp.kls->comment = _eo_tokenizer_token_get(toknz, ( toknz->p)-1);
   }}
	break;
	case 98:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 99:
#line 800 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
   }}
	break;
	case 100:
#line 803 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
   }}
	break;
	case 101:
#line 806 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("  constructors {");
      toknz->current_methods_type = METH_CONSTRUCTOR;
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting++;
      { toknz->cs = 339; goto _again;}
   }}
	break;
	case 102:
#line 814 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("  properties {");
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting++;
      { toknz->cs = 323; goto _again;}
   }}
	break;
	case 103:
#line 821 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      INF("  begin methods");
      toknz->current_methods_type = METH_REGULAR;
      toknz->tmp.fscope = FUNC_PUBLIC;
      toknz->current_nesting++;
      { toknz->cs = 339; goto _again;}
   }}
	break;
	case 104:
#line 829 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      INF("end class: %s", toknz->tmp.kls->name);
      toknz->classes = eina_list_append(toknz->classes, toknz->tmp.kls);
      toknz->tmp.kls = NULL;
      toknz->current_nesting--;
      { toknz->cs = 292; goto _again;}
   }}
	break;
	case 105:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 106:
#line 910 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 107:
#line 913 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 108:
#line 914 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 109:
#line 915 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 110:
#line 800 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
   }}
	break;
	case 111:
#line 803 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
   }}
	break;
	case 112:
#line 829 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      INF("end class: %s", toknz->tmp.kls->name);
      toknz->classes = eina_list_append(toknz->classes, toknz->tmp.kls);
      toknz->tmp.kls = NULL;
      toknz->current_nesting--;
      { toknz->cs = 292; goto _again;}
   }}
	break;
	case 113:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 114:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 115:
#line 386 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("comment[%d] line%03d:%03d", toknz->cs,
          toknz->saved.line, toknz->current_line);
   }}
	break;
	case 116:
#line 927 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      if (!toknz->tmp.kls) ABORT(toknz, "No class!!!");
      INF("begin class: %s", toknz->tmp.kls->name);
      toknz->current_nesting++;
      { toknz->cs = 345; goto _again;}
   }}
	break;
	case 117:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p)+1;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 118:
#line 962 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;}
	break;
	case 119:
#line 395 "lib/eolian/eo_lexer.rl"
	{ toknz->te = ( toknz->p);( toknz->p)--;{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
	case 120:
#line 395 "lib/eolian/eo_lexer.rl"
	{{( toknz->p) = (( toknz->te))-1;}{
      DBG("error[%d]", toknz->cs);
      char *s, *d;
      char buf[BUFSIZE];
      for (s = ( toknz->p), d = buf; (s <= toknz->pe); s++)
        {
           if ((*s == '\r') || (*s == '\n'))
             break;
           *d++ = *s;
        }
      *d = '\0';
      ERR("error n:%d l:%d c:'%c': %s",
          toknz->current_nesting, toknz->current_line, *( toknz->p), buf);
      toknz->cs = eo_tokenizer_error;
      {( toknz->p)++; goto _out; }  /* necessary to stop scanners */
   }}
	break;
#line 4102 "lib/eolian/eo_lexer.c"
		}
	}

_again:
	_acts = _eo_tokenizer_actions + _eo_tokenizer_to_state_actions[ toknz->cs];
	_nacts = (unsigned int) *_acts++;
	while ( _nacts-- > 0 ) {
		switch ( *_acts++ ) {
	case 33:
#line 1 "NONE"
	{ toknz->ts = 0;}
	break;
#line 4115 "lib/eolian/eo_lexer.c"
		}
	}

	if ( ++( toknz->p) != ( toknz->pe) )
		goto _resume;
	_test_eof: {}
	if ( ( toknz->p) == ( toknz->eof) )
	{
	if ( _eo_tokenizer_eof_trans[ toknz->cs] > 0 ) {
		_trans = _eo_tokenizer_eof_trans[ toknz->cs] - 1;
		goto _eof_trans;
	}
	}

	_out: {}
	}

#line 1071 "lib/eolian/eo_lexer.rl"

   if ( toknz->cs == 
#line 4136 "lib/eolian/eo_lexer.c"
-1
#line 1072 "lib/eolian/eo_lexer.rl"
 )
     {
        ERR("%s: wrong termination", source);
        ret = EINA_FALSE;
     }

   return ret;
}

Eo_Tokenizer*
eo_tokenizer_get(void)
{
   Eo_Tokenizer *toknz = calloc(1, sizeof(Eo_Tokenizer));
   if (!toknz) return NULL;

   toknz->ts = NULL;
   toknz->te = NULL;
   /* toknz->top = 0; */
   toknz->source = NULL;
   toknz->max_nesting = 10;
   toknz->current_line = 1;
   toknz->current_nesting = 0;
   toknz->current_methods_type = METH_TYPE_LAST;
   toknz->saved.tok = NULL;
   toknz->saved.line = 0;
   toknz->classes = NULL;

   return toknz;
}

static char *_accessor_type_str[ACCESSOR_TYPE_LAST] = { "setter", "getter" };
static char *_param_way_str[PARAM_WAY_LAST] = { "IN", "OUT", "INOUT" };

void
eo_tokenizer_dump(Eo_Tokenizer *toknz)
{
   const char *s;
   Eina_List *k, *l, *m;

   Eo_Class_Def *kls;
   Eo_Property_Def *prop;
   Eo_Method_Def *meth;
   Eo_Param_Def *param;
   Eo_Accessor_Def *accessor;
   Eo_Event_Def *sgn;
   /* Eo_Ret_Def *ret; */

   EINA_LIST_FOREACH(toknz->classes, k, kls)
     {
        printf("Class: %s (%s)\n",
               kls->name, (kls->comment ? kls->comment : "-"));
        printf("  inherits from :");
        EINA_LIST_FOREACH(kls->inherits, l, s)
           printf(" %s", s);
        printf("\n");
        printf("  implements:");
        EINA_LIST_FOREACH(kls->implements, l, s)
           printf(" %s", s);
        printf("\n");
        printf("  events:\n");
        EINA_LIST_FOREACH(kls->events, l, sgn)
           printf("    %s <%s> (%s)\n", sgn->name, sgn->type, sgn->comment);

        EINA_LIST_FOREACH(kls->constructors, l, meth)
          {
             printf("  constructors: %s\n", meth->name);
             if (meth->ret)
                printf("    return: %s (%s)\n", meth->ret->type, meth->ret->comment);
             printf("    legacy : %s\n", meth->legacy);
             EINA_LIST_FOREACH(meth->params, m, param)
               {
                  printf("    param: %s %s : %s (%s)\n",
                         _param_way_str[param->way], param->name,
                         param->type, param->comment);
               }
          }

        EINA_LIST_FOREACH(kls->properties, l, prop)
          {
             printf("  property: %s\n", prop->name);
             EINA_LIST_FOREACH(prop->keys, m, param)
               {
                  printf("    key: %s : %s (%s)\n",
                         param->name, param->type, param->comment);
               }
             EINA_LIST_FOREACH(prop->values, m, param)
               {
                  printf("    value: %s : %s (%s)\n",
                         param->name, param->type, param->comment);
               }
             EINA_LIST_FOREACH(prop->accessors, m, accessor)
               {
                  printf("    accessor: %s : %s (%s)\n",
                         (accessor->ret?accessor->ret->type:""),
                         _accessor_type_str[accessor->type],
                         accessor->comment);
                  printf("      legacy : %s\n", accessor->legacy);
               }
          }

        EINA_LIST_FOREACH(kls->methods, l, meth)
          {
             printf("  method: %s\n", meth->name);
             if (meth->ret)
                printf("    return: %s (%s)\n", meth->ret->type, meth->ret->comment);
             printf("    legacy : %s\n", meth->legacy);
             printf("    obj_const : %s\n", meth->obj_const?"true":"false");
             EINA_LIST_FOREACH(meth->params, m, param)
               {
                  printf("    param: %s %s : %s (%s)\n",
                         _param_way_str[param->way], param->name,
                         param->type, param->comment);
               }
          }

     }

}

static Eina_Inlist *
_types_extract(const char *buf, int len)
{
   const char *save_buf = buf;
   Eolian_Type types = NULL;
   long depth = 0;
   char *tmp_type = malloc(2 * len + 1);

   while (len > 0)
     {
        char *d = tmp_type;
        Eina_Bool end_type = EINA_FALSE;
        Eina_Bool is_own = EINA_FALSE;
        char c;
        Eina_Bool in_spaces = EINA_TRUE, in_stars = EINA_FALSE;
        while (len > 0 && !end_type)
          {
             switch (c = *buf++)
               {
                /* @own */
                case '@':
                     {
                        if (!strncmp(buf, "own ", 4))
                          {
                             is_own = EINA_TRUE;
                             buf += 4; len -= 4;
                          }
                        break;
                     }
                /* if '*', we have to add a space. We set in_spaces to true in
                 * case spaces are between stars, to be sure we remove them.
                 */
                case '*':
                     {
                        if (!in_stars && !in_spaces)
                          {
                             *d++ = ' ';
                             in_stars = EINA_TRUE;
                             in_spaces = EINA_TRUE;
                          }
                        *d++ = c;
                        break;
                     }
                /* Only the first space is inserted. */
                case ' ':
                     {
                        if (!in_spaces) *d++ = c;
                        in_spaces = EINA_TRUE;
                        break;
                     }
                case '<':
                     {
                        if (depth < 0)
                          {
                             ERR("%s: Cannot reopen < after >", save_buf);
                             goto error;
                          }
                        depth++;
                        end_type = EINA_TRUE;
                        break;
                     }
                case '>':
                     {
                        if (depth == 0)
                          {
                             ERR("%s: Too much >", save_buf);
                             goto error;
                          }
                        if (depth > 0 && d == tmp_type)
                          {
                             ERR("%s: empty type inside <>", save_buf);
                             goto error;
                          }
                        if (depth > 0) depth *= -1;
                        depth++;
                        end_type = EINA_TRUE;
                        break;
                     }
                default:
                     {
                        *d++ = c;
                        in_spaces = EINA_FALSE;
                        in_stars = EINA_FALSE;
                     }
               }
             len--;
          }
        if (d != tmp_type)
          {
             *d = '\0';
             types = database_type_append(types, tmp_type, is_own);
          }
     }
   if (depth)
     {
        ERR("%s: < and > are not well used.", save_buf);
        goto error;
     }
   goto success;
error:
   database_type_del(types);
   types = NULL;
success:
   free(tmp_type);
   return types;
}

Eina_Bool
eo_tokenizer_database_fill(const char *filename)
{
   Eina_Bool ret = EINA_FALSE;
   const char *s;
   Eina_List *k, *l, *m;

   Eo_Class_Def *kls;
   Eo_Property_Def *prop;
   Eo_Method_Def *meth;
   Eo_Param_Def *param;
   Eo_Accessor_Def *accessor;
   Eo_Event_Def *event;
   Eo_Implement_Def *impl;

   FILE *stream = NULL;
   char *buffer = NULL;

   Eo_Tokenizer *toknz = eo_tokenizer_get();
   if (!toknz)
     {
        ERR("can't create eo_tokenizer");
        goto end;
     }

   stream = fopen(filename, "rb");
   if (!stream)
     {
        ERR("unable to read in %s", filename);
        goto end;
     }

   buffer = malloc(BUFSIZE);
   if (!buffer)
     {
        ERR("unable to allocate read buffer");
        goto end;
     }

   unsigned int len = fread(buffer, 1, BUFSIZE, stream);

   if (!len)
     {
        ERR("%s: is an empty file", filename);
        goto end;
     }

   if (len == BUFSIZE)
     {
        ERR("%s: buffer(%d) is full, might not be big enough.", filename, len);
        goto end;
     }

   if (!eo_tokenizer_mem_walk(toknz, filename, buffer, len)) goto end;

   if (!toknz->classes)
     {
        ERR("No classes for file %s", filename);
        goto end;
     }

   EINA_LIST_FOREACH(toknz->classes, k, kls)
     {
        database_class_add(kls->name, kls->type);
        database_class_file_set(kls->name, filename);

        if (kls->comment) database_class_description_set(kls->name, kls->comment);

        EINA_LIST_FOREACH(kls->inherits, l, s)
           database_class_inherit_add(kls->name, s);

        if (kls->legacy_prefix)
          {
             database_class_legacy_prefix_set(kls->name, kls->legacy_prefix);
          }
        if (kls->eo_prefix)
          {
             database_class_eo_prefix_set(kls->name, kls->eo_prefix);
          }
        if (kls->data_type)
          {
             database_class_data_type_set(kls->name, kls->data_type);
          }
        EINA_LIST_FOREACH(kls->constructors, l, meth)
          {
             Eolian_Function foo_id = database_function_new(meth->name, EOLIAN_CTOR);
             database_class_function_add(kls->name, foo_id);
             if (meth->ret) database_function_return_comment_set(foo_id, EOLIAN_METHOD, meth->ret->comment);
             database_function_data_set(foo_id, EOLIAN_LEGACY, meth->legacy);
             EINA_LIST_FOREACH(meth->params, m, param)
               {
                  Eolian_Type type = _types_extract(param->type, strlen(param->type));
                  database_method_parameter_add(foo_id, (Eolian_Parameter_Dir)param->way, type, param->name, param->comment);
               }
          }

        EINA_LIST_FOREACH(kls->properties, l, prop)
          {
             Eolian_Function foo_id = database_function_new(prop->name, EOLIAN_UNRESOLVED);
             database_function_scope_set(foo_id, prop->scope);
             EINA_LIST_FOREACH(prop->keys, m, param)
               {
                  Eolian_Type type = _types_extract(param->type, strlen(param->type));
                  Eolian_Function_Parameter p = database_property_key_add(
                        foo_id, type, param->name, param->comment);
                  database_parameter_nonull_set(p, param->nonull);
               }
             EINA_LIST_FOREACH(prop->values, m, param)
               {
                  Eolian_Type type = _types_extract(param->type, strlen(param->type));
                  Eolian_Function_Parameter p = database_property_value_add(
                        foo_id, type, param->name, param->comment);
                  database_parameter_nonull_set(p, param->nonull);
               }
             EINA_LIST_FOREACH(prop->accessors, m, accessor)
               {
                  database_function_type_set(foo_id, (accessor->type == SETTER?EOLIAN_PROP_SET:EOLIAN_PROP_GET));
                  if (accessor->ret && accessor->ret->type)
                    {
                       Eolian_Function_Type ftype =
                          accessor->type == SETTER?EOLIAN_PROP_SET:EOLIAN_PROP_GET;
                       Eolian_Type types = _types_extract(accessor->ret->type, strlen(accessor->ret->type));
                       database_function_return_type_set(foo_id, ftype, types);
                       database_function_return_comment_set(foo_id,
                             ftype, accessor->ret->comment);
                       database_function_return_flag_set_as_warn_unused(foo_id,
                             ftype, accessor->ret->warn_unused);
                       database_function_return_dflt_val_set(foo_id,
                             ftype, accessor->ret->dflt_ret_val);
                    }
                  if (accessor->legacy)
                    {
                       database_function_data_set(foo_id,
                             (accessor->type == SETTER?EOLIAN_LEGACY_SET:EOLIAN_LEGACY_GET),
                             accessor->legacy);
                    }
                  database_function_description_set(foo_id,
                        (accessor->type == SETTER?EOLIAN_COMMENT_SET:EOLIAN_COMMENT_GET),
                        accessor->comment);
                  Eo_Accessor_Param *acc_param;
                  Eina_List *m2;
                  EINA_LIST_FOREACH(accessor->params, m2, acc_param)
                    {
                       Eolian_Function_Parameter desc = eolian_function_parameter_get(foo_id, acc_param->name);
                       if (!desc)
                         {
                            printf("Error - %s not known as parameter of property %s\n", acc_param->name, prop->name);
                         }
                       else
                          if (strstr(acc_param->attrs, "const"))
                            {
                               database_parameter_const_attribute_set(desc, accessor->type == GETTER, EINA_TRUE);
                            }
                    }
               }
             database_class_function_add(kls->name, foo_id);
          }

        EINA_LIST_FOREACH(kls->methods, l, meth)
          {
             Eolian_Function foo_id = database_function_new(meth->name, EOLIAN_METHOD);
             database_function_scope_set(foo_id, meth->scope);
             database_class_function_add(kls->name, foo_id);
             if (meth->ret)
               {
                  Eolian_Type types = _types_extract(meth->ret->type, strlen(meth->ret->type));
                  database_function_return_type_set(foo_id, EOLIAN_METHOD, types);
                  database_function_return_comment_set(foo_id, EOLIAN_METHOD, meth->ret->comment);
                  database_function_return_flag_set_as_warn_unused(foo_id,
                        EOLIAN_METHOD, meth->ret->warn_unused);
                  database_function_return_dflt_val_set(foo_id,
                        EOLIAN_METHOD, meth->ret->dflt_ret_val);
               }
             database_function_description_set(foo_id, EOLIAN_COMMENT, meth->comment);
             database_function_data_set(foo_id, EOLIAN_LEGACY, meth->legacy);
             database_function_object_set_as_const(foo_id, meth->obj_const);
             EINA_LIST_FOREACH(meth->params, m, param)
               {
                  Eolian_Type type = _types_extract(param->type, strlen(param->type));
                  Eolian_Function_Parameter p = database_method_parameter_add(foo_id,
                        (Eolian_Parameter_Dir)param->way, type, param->name, param->comment);
                  database_parameter_nonull_set(p, param->nonull);
               }
          }

        EINA_LIST_FOREACH(kls->implements, l, impl)
          {
             const char *class = impl->meth_name;
             Eina_Bool virtual_pure = EINA_FALSE;
             if (!strcmp(class, "class::constructor"))
               {
                  database_class_ctor_enable_set(kls->name, EINA_TRUE);
                  continue;
               }
             if (!strcmp(class, "class::destructor"))
               {
                  database_class_dtor_enable_set(kls->name, EINA_TRUE);
                  continue;
               }
             if (!strncmp(class, "virtual::", 9)) virtual_pure = EINA_TRUE;
             char *func = strstr(class, "::");
             if (func) *func = '\0';
             func += 2;
             Eolian_Function_Type ftype = EOLIAN_UNRESOLVED;
             char *type_as_str = strstr(func, "::");
             if (type_as_str)
               {
                  *type_as_str = '\0';
                  if (!strcmp(type_as_str+2, "set")) ftype = EOLIAN_PROP_SET;
                  else if (!strcmp(type_as_str+2, "get")) ftype = EOLIAN_PROP_GET;
               }
             if (virtual_pure)
               {
                  /* Search the function into the existing functions of the current class */
                  Eolian_Function foo_id = eolian_class_function_find_by_name(
                        kls->name, func, ftype);
                  if (!foo_id)
                    {
                       ERR("Error - %s not known in class %s", class + 9, kls->name);
                       goto end;
                    }
                  database_function_set_as_virtual_pure(foo_id, ftype);
                  continue;
               }
             Eolian_Implement impl_desc = database_implement_new(class, func, ftype);
             database_class_implement_add(kls->name, impl_desc);
          }

        EINA_LIST_FOREACH(kls->events, l, event)
          {
             Eolian_Event ev = database_event_new(event->name, event->type, event->comment);
             database_class_event_add(kls->name, ev);
          }

     }

   ret = EINA_TRUE;
end:
   if (buffer) free(buffer);
   if (stream) fclose(stream);
   if (toknz) eo_tokenizer_free(toknz);
   return ret;
}

void
eo_tokenizer_free(Eo_Tokenizer *toknz)
{
   Eo_Class_Def *kls;

   if (toknz->source)
     eina_stringshare_del(toknz->source);

   EINA_LIST_FREE(toknz->classes, kls)
      eo_definitions_class_def_free(kls);

   free(toknz);
}

